import codecs
import difflib
import io
import os
import re
import shutil
import py_compile
import xmlschema
from pathlib import Path
from tempfile import mkdtemp
from ..utils import _q, _lb, _rb, TESTS_DATA_PATH

from psychopy import experiment
from psychopy.experiment.components.settings import SettingsComponent
from psychopy.experiment.components.unknown import UnknownComponent
from ...scripts import psyexpCompile


class TestExperiment:
    @classmethod
    def setup_class(cls):
        cls.exp = experiment.Experiment() # create once, not every test
        try:
            cls.tempDir = mkdtemp(dir=Path(__file__).root, prefix='psychopy-tests-app')
        except (PermissionError, OSError):
            # can't write to root on Linux
            cls.tempDir = mkdtemp(prefix='psychopy-tests-app')

    def setup(self):
        # Make a basic experiment with one routine
        self.exp = experiment.Experiment()
        self.rt = self.exp.addRoutine("testRoutine")
        self.exp.flow.addRoutine(self.rt, 0)
        # Add one of every component to that routine (default params)
        for compName, compClass in experiment.getAllComponents().items():
            if compClass != SettingsComponent:
                comp = compClass(exp=self.exp, parentName=self.rt.name, name=f"test{compName}")
                self.rt.append(comp)
        # Add one of every standalone routine
        for rtName, rtClass in experiment.getAllStandaloneRoutines().items():
            rt = rtClass(exp=self.exp, name=f"test{rtName}")
            self.exp.addStandaloneRoutine(rt.name, rt)
        # Add all routines to the flow
        for rt in self.exp.routines.values():
            self.exp.flow.addRoutine(rt, 0)

    def teardown_class(self):
        shutil.rmtree(self.tempDir)

    # ---------
    # Utilities

    @staticmethod
    def _checkCompile(py_file):
        # compile the temp file to .pyc, catching error msgs
        # (including no file at all):
        py_file = str(py_file)
        try:
            py_compile.compile(py_file, doraise=True)
        except py_compile.PyCompileError as err:
            err.msg = py_file
            raise err
        return py_file + 'c'

    # ---------
    # Utilities

    def test_add_routine(self):
        exp = experiment.Experiment()

        # Test adding a regular routine
        rt = exp.addRoutine(f"testRoutine")
        # Check that the routine name is present
        assert rt.name in exp.routines
        # Check that the routine is a Routine
        assert isinstance(exp.routines[rt.name], experiment.routines.Routine), (
             f"Routine {rt.name} should be Routine but was {type(exp.routines[rt.name]).__name__}"
        )
        # Test adding standalone routines
        for rtName, rtClass in experiment.getAllStandaloneRoutines().items():
            # Make and add standalone routine of this type
            rt = rtClass(exp=exp, name=f"test{rtClass.__name__}")
            exp.addStandaloneRoutine(rt.name, rt)
            # Check that the routine name is present
            assert rt.name in exp.routines, f"Could not find {rtClass.__name__} in experiment after adding"
            # Check that the routine is a Routine
            assert isinstance(exp.routines[rt.name], rtClass), (
                f"Routine {rt.name} should be {rtClass.__name__} but was {type(exp.routines[rt.name]).__name__}"
            )

        # Check that none of these routines are in the flow yet
        for rtName, rt in exp.routines.items():
            assert rt not in exp.flow, (
                f"Routine {rtName} of type {type(rt).__name__} found in experiment flow before being added"
            )
        # Check that none of these routines appear in the compiled script yet
        pyScript = exp.writeScript(target="PsychoPy")
        jsScript = exp.writeScript(target="PsychoJS")
        for rtName, rt in exp.routines.items():
            if "PsychoPy" in type(rt).targets:
                assert rtName not in pyScript, (
                    f"Routine {rtName} of type {type(rt).__name__} found in Python script before being added to flow"
                )
            if "PsychoJS" in type(rt).targets:
                assert rtName not in jsScript, (
                    f"Routine {rtName} of type {type(rt).__name__} found in JS script before being added to flow"
                )

        # Add routines to flow
        for rtName, rt in exp.routines.items():
            exp.flow.addRoutine(rt, 0)
        # Check that they are in flow now
        for rtName, rt in exp.routines.items():
            assert rt in exp.flow, (
                f"Routine {rtName} of type {type(rt).__name__} not found in experiment flow after being added"
            )
        # Check that all of these routines appear in the compiled script yet
        pyScript = exp.writeScript(target="PsychoPy")
        jsScript = exp.writeScript(target="PsychoJS")
        for rtName, rt in exp.routines.items():
            if "PsychoPy" in type(rt).targets:
                assert rtName in pyScript, (
                    f"Routine {rtName} of type {type(rt).__name__} not found in Python script after being added to flow"
                )
            if "PsychoJS" in type(rt).targets:
                assert rtName in jsScript, (
                    f"Routine {rtName} of type {type(rt).__name__} not found in JS script after being added to flow"
                )

    def test_xml(self):
        isTime = re.compile(r"\d+:\d+(:\d+)?( [AP]M)?")
        # Get all psyexp files in demos folder
        demosFolder = Path(self.exp.prefsPaths['demos']) / 'builder'
        for file in demosFolder.glob("**/*.psyexp"):
            # Create experiment and load from psyexp
            exp = experiment.Experiment()
            exp.loadFromXML(file)
            # Compile to get what script should look like
            target = exp.writeScript()
            # Save as XML
            temp = str(Path(self.tempDir) / "testXML.psyexp")
            exp.saveToXML(temp)
            # Load again
            exp.loadFromXML(temp)
            # Compile again
            test = exp.writeScript()
            # Remove any timestamps from script (these can cause false errors if compile takes longer than a second)
            test = re.sub(isTime, "", test)
            target = re.sub(isTime, "", target)
            # Compare two scripts to make sure saving and loading hasn't changed anything
            diff = difflib.unified_diff(target.splitlines(), test.splitlines())
            assert list(diff) == []

    def test_xsd(self):
        # get files

        psyexp_files = []

        for root, dirs, files in os.walk(os.path.join(self.exp.prefsPaths['demos'], 'builder')):
            for f in files:
                if f.endswith('.psyexp'):
                    psyexp_files.append(os.path.join(root, f))

        # get schema

        schema_name = Path(self.exp.prefsPaths['psychopy']) / 'experiment' / 'experiment.xsd'
        schema = xmlschema.XMLSchema(str(schema_name))

        for psyexp_file in psyexp_files:
            assert schema.is_valid(psyexp_file), (
                    f"Error in {psyexp_file}:\n" + "\n".join(err.reason for err in schema.iter_errors(psyexp_file))
            )

    def test_problem_experiments(self):
        """
        Tests that some known troublemaker experiments compile as intended. Cases are structured like so:
        file : `.psyexp` file for the experiment
        comparison : Type of comparison to do, can be any of:
            - contains : Compiled script should contain the specified answer
            - excludes : Compiled script should NOT contain the specified answer
            - equals : Compiled script should match the answer
        ans : The string to do specified comparison against
        """
        # Define some cases
        tykes = [
            {'file': Path(TESTS_DATA_PATH) / "retroListParam.psyexp", 'comparison': "contains",
             'ans': f"{_lb}{_q}left{_q}, ?{_q}down{_q}, ?{_q}right{_q}{_rb}"}
        ]
        # Temp outfile to use
        outfile = Path(self.tempDir) / 'outfile.py'
        # Run each case
        for case in tykes:
            # Compile experiment
            psyexpCompile.compileScript(infile=case['file'], outfile=str(outfile))
            # Get compiled script as a string
            with open(outfile, "r") as f:
                outscript = f.read()
            # Do comparison
            if case['comparison'] == "contains":
                assert re.search(case['ans'], outscript), (
                    f"No match found for `{case['ans']}` in compile of {case['file'].name}. View compile here: {outfile}"
                )
            if case['comparison'] == "excludes":
                assert not re.search(case['ans'], outscript), (
                    f"Unwanted match found for `{case['ans']}` in compile of {case['file'].name}. View compile here: {outfile}"
                )
            if case['comparison'] == "equals":
                assert re.fullmatch(case['ans'], outscript), (
                    f"Compile of {case['file'].name} did not match {case['ans']}. View compile here: {outfile}"
                )

    def test_future(self):
        """An experiment file with made-up params and routines to see whether
        future versions of experiments will get loaded.
        """
        # Define some component names which should be UnknownComponent
        invComps = [
            "Bomber",
            "Banana"
        ]
        # Define some component names which should be known
        validComps = [
            "ISI",
            "text",
            "sound_1",
            "buttonBox",
        ]
        # Define some param names which should be InvalidCtrl
        invParams = [
            "Wacky",
        ]
        # Define some param names which should be known
        validParams = [
            "pos",
            "size",
            "color",
        ]

        # Load experiment from file
        expfile = Path(self.exp.prefsPaths['tests']) / 'data' / 'futureParams.psyexp'
        self.exp.loadFromXML(expfile) # reload the edited file
        # Iterate through components to check recognition of component class and params
        for comp in self.exp.routines['trial']:
            if comp.name in invComps:
                # If component should be unknown, make sure it is
                assert isinstance(comp, UnknownComponent), (
                    f"Component {comp.name} is a made-up component and so should not be recognised by PsychoPy, but "
                    f"was recognised as {type(comp).__name__}."
                )
                # Iterate through all params in invalid component to check they aren't recognised
                for paramName, param in comp.params.items():
                    if paramName in list(UnknownComponent(self.exp, 'trial').params):
                        # If param name is defined in UnknownComponent, it is always valid
                        assert param.inputType != "inv", (
                            f"Parameter {paramName} is defined in UnknownComponent, so should always be valid, but in "
                            f"component {comp.name} it was not recognised."
                        )
                    else:
                        # Otherwise, it should be invalid
                        assert param.inputType == "inv", (
                            f"In an UnknownComponent, all params save for those defined in UnknownComponent should be "
                            f"invalid."
                        )
            if comp.name in validComps:
                # If component should be known, make sure it is
                assert not isinstance(comp, UnknownComponent), (
                    f"Component {comp.name} should be recognised by PsychoPy, but was interpreted as UnknownComponent."
                )
                # Iterate through all params in valid component to check recognition
                for paramName, param in comp.params.items():
                    if paramName in invParams:
                        # If param should be unknown, make sure it is
                        assert param.inputType == "inv", (
                            f"Param {paramName} of {comp.name} is a made-up param and so its input type should be `inv`, "
                            f"but instead it is {param.inputType}"
                        )
                    if paramName in validParams:
                        # If param should be known, make sure it is
                        assert param.inputType != "inv", (
                            f"Param {paramName} of {comp.name} should be known to PsychoPy, but its inputType was marked "
                            f"as `inv`"
                        )

        # Make sure it builds
        script = self.exp.writeScript(expPath=expfile)
        py_file = Path(self.tempDir) / 'testFutureFile.py'
        # Save script
        with codecs.open(py_file, 'w', 'utf-8-sig') as f:
            f.write(script)

        # Check it compiles to pyc
        self._checkCompile(py_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""A Backend class defines the core low-level functions required by a Window
class, such as the ability to create an OpenGL context and flip the window.

Users simply call visual.Window(..., winType='pyglet') and the winType is then
used by backends.getBackend(winType) which will locate the appropriate class
and initialize an instance using the attributes of the Window.
"""

import os
import sys
import pygame
from abc import abstractmethod
from ._base import BaseBackend
import psychopy
from psychopy import core
from psychopy.tools.attributetools import attributeSetter

try:
    import pyglet
    GL = pyglet.gl
except ImportError:
    import OpenGL
    GL = OpenGL


class PygameBackend(BaseBackend):
    """The pygame backend is built on SDL for cross-platform controls
    """
    GL = GL
    winTypeName = 'pygame'

    def __init__(self, win, backendConf=None):
        """Set up the backend window according the params of the PsychoPy win

        Before PsychoPy 1.90.0 this code was executed in Window._setupPygame()

        Parameters
        ----------
        win : `psychopy.visual.Window` instance
            PsychoPy Window (usually not fully created yet).
        backendConf : `dict` or `None`
            Backend configuration options. Options are specified as a dictionary
            where keys are option names and values are settings. This backend
            currently takes no additional settings.

        Examples
        --------
        Create a window using the Pygame backend::

            import psychopy.visual as visual
            win = visual.Window(winType='glfw', backendOptions=options)

        """
        BaseBackend.__init__(self, win)  # sets up self.win=win as weakref

        # pygame.mixer.pre_init(22050,16,2)  # set the values to initialise
        # sound system if it gets used
        pygame.init()
        if win.allowStencil:
            pygame.display.gl_set_attribute(pygame.locals.GL_STENCIL_SIZE, 8)

        try:  # to load an icon for the window
            iconFile = os.path.join(psychopy.__path__[0], 'psychopy.png')
            icon = pygame.image.load(iconFile)
            pygame.display.set_icon(icon)
        except Exception:
            pass  # doesn't matter
        win.useRetina = False
        # these are ints stored in pygame.locals
        winSettings = pygame.OPENGL | pygame.DOUBLEBUF
        if win._isFullScr:
            winSettings = winSettings | pygame.FULLSCREEN
            # check screen size if full screen
            scrInfo = pygame.display.Info()
            win._checkMatchingSizes(win.clientSize,
                                    [scrInfo.current_w,
                                     scrInfo.current_h])
        elif not win.pos:
            # centre video
            os.environ['SDL_VIDEO_CENTERED'] = "1"
        else:
            os.environ['SDL_VIDEO_WINDOW_POS'] = '%i,%i' % (win.pos[0],
                                                            win.pos[1])
        if sys.platform == 'win32':
            os.environ['SDL_VIDEODRIVER'] = 'windib'
        if not win.allowGUI:
            winSettings = winSettings | pygame.NOFRAME
            self.setMouseVisibility(False)
            pygame.display.set_caption('PsychoPy (NB use with allowGUI=False '
                                       'when running properly)')
        else:
            self.setMouseVisibility(True)
            pygame.display.set_caption('PsychoPy')
        self.winHandle = pygame.display.set_mode(win.size.astype('i'),
                                                 winSettings)
        self._frameBufferSize = win.clientSize
        pygame.display.set_gamma(1.0)  # this will be set appropriately later
        # This is causing segfault although it used to be for pyglet only anyway
        # pygame under mac is not syncing to refresh although docs say it should
        # if sys.platform == 'darwin':
        #     platform_specific.syncSwapBuffers(2)

    @property
    def frameBufferSize(self):
        """Framebuffer size (w, h)."""
        return self._frameBufferSize

    def swapBuffers(self, flipThisFrame=True):
        """Do the actual flipping of the buffers (Window will take care of
        additional things like timestamping. Keep this methods as short as poss

        :param flipThisFrame: has no effect on this backend
        """
        if pygame.display.get_init():
            if flipThisFrame:
                pygame.display.flip()
            # keeps us in synch with system event queue
            self.dispatchEvents()
        else:
            core.quit()  # we've unitialised pygame so quit

    def close(self):
        """Close the window and uninitialize the resources
        """
        pygame.display.quit()

    @property
    def shadersSupported(self):
        """This is a read-only property indicating whether or not this backend
        supports OpenGL shaders"""
        return False

    def setMouseVisibility(self, visibility):
        """Set visibility of the mouse to True or False"""
        pygame.mouse.set_visible(visibility)

    # Optional, depending on backend needs

    def dispatchEvents(self):
        """This method is not needed for all backends but for engines with an
        event loop it may be needed to pump for new events (e.g. pyglet)
        """
        pygame.event.pump()

    def onResize(self, width, height):
        """This does nothing; not supported by our pygame backend at the moment
        """
        pass  # the pygame window doesn't currently support resizing

    @attributeSetter
    def gamma(self, gamma):
        self.__dict__['gamma'] = gamma
        # use pygame's own function for this
        pygame.display.set_gamma(gamma[0], gamma[1], gamma[2])

    @attributeSetter
    def gammaRamp(self, gammaRamp):
        """Gets the gamma ramp or sets it to a new value (an Nx3 or Nx1 array)
        """
        self.__dict__['gammaRamp'] = gammaRamp
        # use pygame's own function for this
        pygame.display.set_gamma_ramp(
                gammaRamp[:, 0], gammaRamp[:, 1], gammaRamp[:, 2])

    def setFullScr(self, value):
        """Sets the window to/from full-screen mode"""
        raise NotImplementedError("Toggling fullscreen mode is not currently "
                             "supported on pygame windows")

    # --------------------------------------------------------------------------
    # Mouse related methods (e.g., event handlers)
    #
    # These methods are used to handle mouse events. Each function is bound to
    # the appropriate callback which registers the mouse event with the global
    # mouse event handler (psychopy.hardware.mouse.Mouse). Each callback has an
    # `*args` parameter which allows the backend to pass whatever parameters.
    #

    @abstractmethod
    def onMouseButton(self, *args, **kwargs):
        """Event handler for any mouse button event (pressed and released).

        This is used by backends which combine both button state changes into
        a single event. Usually this would pass events to the appropriate
        `onMouseButtonPress` and `onMouseButtonRelease` methods.
        """
        raise NotImplementedError(
            "`onMouseButton` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseButtonPress(self, *args, **kwargs):
        """Event handler for mouse press events. This handler can also be used
        for release events if the backend passes all button events to the same
        callback.
        """
        raise NotImplementedError(
            "`onMouseButtonPress` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseButtonRelease(self, *args, **kwargs):
        """Event handler for mouse release events."""
        raise NotImplementedError(
            "`onMouseButtonRelease` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseScroll(self, *args, **kwargs):
        """Event handler for mouse scroll events. Called when the mouse scroll
        wheel is moved."""
        raise NotImplementedError(
            "`onMouseScroll` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseMove(self, *args, **kwargs):
        """Event handler for mouse move events."""
        raise NotImplementedError(
            "`onMouseMove` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseEnter(self, *args, **kwargs):
        """Event called when the mouse enters the window. Some backends might
        combine enter and leave events to the same callback, this will handle
        both if so.
        """
        raise NotImplementedError(
            "`onMouseEnter` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseLeave(self, *args, **kwargs):
        """Event called when a mouse leaves the window."""
        raise NotImplementedError(
            "`onMouseLeave` is not yet implemented for this backend.")

    @abstractmethod
    def getMousePos(self):
        """Get the position of the mouse on the current window.

        Returns
        -------
        ndarray
            Position `(x, y)` in window coordinates.

        """
        raise NotImplementedError(
            "`getMousePos` is not yet implemented for this backend.")

    @abstractmethod
    def setMousePos(self, pos):
        """Set/move the position of the mouse on the current window.

        Parameters
        ----------
        pos : ArrayLike
            Position `(x, y)` in window coordinates.

        """
        raise NotImplementedError(
            "`setMousePos` is not yet implemented for this backend.")

    def setMouseType(self, name='arrow'):
        """Change the appearance of the cursor for this window. Cursor types
        provide contextual hints about how to interact with on-screen objects.

        **Deprecated!** Use `setMouseCursor` instead.

        Parameters
        ----------
        name : str
            Type of standard cursor to use.

        """
        self.setMouseCursor(name)

    @abstractmethod
    def setMouseCursor(self, cursorType='default'):
        """Change the appearance of the cursor for this window. Cursor types
        provide contextual hints about how to interact with on-screen objects.

        The graphics used 'standard cursors' provided by the operating system.
        They may vary in appearance and hot spot location across platforms. The
        following names are valid on most platforms:

        * ``arrow`` or ``default`` : Default system pointer.
        * ``ibeam`` or ``text`` : Indicates text can be edited.
        * ``crosshair`` : Crosshair with hot-spot at center.
        * ``hand`` : A pointing hand.
        * ``hresize`` : Double arrows pointing horizontally.
        * ``vresize`` : Double arrows pointing vertically.
        * ``help`` : Arrow with a question mark beside it (Windows only).
        * ``no`` : 'No entry' sign or circle with diagonal bar.
        * ``size`` : Vertical and horizontal sizing.
        * ``downleft`` or ``upright`` : Double arrows pointing diagonally with
          positive slope (Windows only).
        * ``downright`` or ``upleft`` : Double arrows pointing diagonally with
          negative slope (Windows only).
        * ``lresize`` : Arrow pointing left (Mac OS X only).
        * ``rresize`` : Arrow pointing right (Mac OS X only).
        * ``uresize`` : Arrow pointing up (Mac OS X only).
        * ``dresize`` : Arrow pointing down (Mac OS X only).
        * ``wait`` : Hourglass (Windows) or watch (Mac OS X) to indicate the
           system is busy.
        * ``waitarrow`` : Hourglass beside a default pointer (Windows only).

        In cases where a cursor is not supported, the default for the system
        will be used.

        Parameters
        ----------
        cursorType : str
            Type of standard cursor to use. If not specified, `'default'` is
            used.

        Notes
        -----
        * On some platforms the 'crosshair' cursor may not be visible on uniform
          grey backgrounds.

        """
        raise NotImplementedError(
            "`setMouseCursor` is not yet implemented for this backend.")

    @abstractmethod
    def setMouseExclusive(self, exclusive):
        """Set mouse exclusivity.

        Parameters
        ----------
        exclusive : bool
            Mouse exclusivity mode.

        """
        raise NotImplementedError(
            "`setMouseExclusive` is not yet implemented for this backend.")


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to image handling"""

try:
    from PIL import Image
except ImportError:
    import Image

import numpy

from psychopy.tools.typetools import float_uint8


def array2image(a):
    """Takes an array and returns an image object (PIL)."""
    # fredrik lundh, october 1998
    #
    # fredrik@pythonware.com
    # http://www.pythonware.com
    #
    if a.dtype.kind in ['u', 'I', 'B']:
        mode = "L"
    elif a.dtype.kind in [numpy.float32, 'f']:
        mode = "F"
    else:
        raise ValueError("unsupported image mode")

    im = Image.frombytes(mode, (a.shape[1], a.shape[0]), a.tobytes())

    return im


def image2array(im):
    """Takes an image object (PIL) and returns a numpy array.
    """
    #     fredrik lundh, october 1998
    #
    #     fredrik@pythonware.com
    #     http://www.pythonware.com
    #
    if im.mode not in ("L", "F"):
        raise ValueError("can only convert single-layer images")

    imdata = im.tobytes()

    if im.mode == "L":
        a = numpy.frombuffer(imdata, numpy.uint8)
    else:
        a = numpy.frombuffer(imdata, numpy.float32)

    a.shape = im.size[1], im.size[0]
    return a


def makeImageAuto(inarray):
    """Combines float_uint8 and image2array operations ie. scales a numeric
    array from -1:1 to 0:255 and converts to PIL image format.
    """
    return image2array(float_uint8(inarray))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A class for getting numeric or categorical ratings, e.g., a 1-to-7 scale."""

# Part of the PsychoPy library
# Copyright (C) 2015 Jonathan Peirce
# Distributed under the terms of the GNU General Public License (GPL).




import copy
import numpy as np

from psychopy import core, logging, event, layout
from psychopy.tools import arraytools
from .basevisual import MinimalStim, WindowMixin, ColorMixin, BaseVisualStim
from .rect import Rect
from .grating import GratingStim
from .elementarray import ElementArrayStim
from .circle import Circle
from .shape import ShapeStim
from . import TextBox2
from ..tools.attributetools import logAttrib, setAttribute, attributeSetter
from ..constants import FINISHED, STARTED, NOT_STARTED

# Set to True to make borders visible for debugging
debug = False


class Slider(MinimalStim, WindowMixin, ColorMixin):
    """A class for obtaining ratings, e.g., on a 1-to-7 or categorical scale.

    A simpler alternative to RatingScale, to be customised with code rather
    than with arguments.

    A RatingScale instance is a re-usable visual object having a ``draw()``
    method, with customizable appearance and response options. ``draw()``
    displays the rating scale, handles the subject's mouse or key responses,
    and updates the display. When the subject accepts a selection,
    ``.noResponse`` goes ``False`` (i.e., there is a response).

    You can call the ``getRating()`` method anytime to get a rating,
    ``getRT()`` to get the decision time, or ``getHistory()`` to obtain
    the entire set of (rating, RT) pairs.

    For other examples see Coder Demos -> stimuli -> ratingsNew.py.

    :Authors:
        - 2018: Jon Peirce
    """

    def __init__(self,
                 win,
                 ticks=(1, 2, 3, 4, 5),
                 labels=None,
                 startValue=None,
                 pos=(0, 0),
                 size=None,
                 units=None,
                 flip=False,
                 ori=0,
                 style='rating', styleTweaks=[],
                 granularity=0,
                 readOnly=False,
                 labelColor='White',
                 markerColor='Red',
                 lineColor='White',
                 colorSpace='rgb',
                 opacity=None,
                 font='Helvetica Bold',
                 depth=0,
                 name=None,
                 labelHeight=None,
                 labelWrapWidth=None,
                 autoDraw=False,
                 autoLog=True,
                 # Synonyms
                 color=False,
                 fillColor=False,
                 borderColor=False):
        """

        Parameters
        ----------
        win : psychopy.visual.Window
            Into which the scale will be rendered

        ticks : list or tuple
            A set of values for tick locations. If given a list of numbers then
            these determine the locations of the ticks (the first and last
            determine the endpoints and the rest are spaced according to
            their values between these endpoints.

        labels : a list or tuple
            The text to go with each tick (or spaced evenly across the ticks).
            If you give 3 labels but 5 tick locations then the end and middle
            ticks will be given labels. If the labels can't be distributed
            across the ticks then an error will be raised. If you want an
            uneven distribution you should include a list matching the length
            of ticks but with some values set to None

        pos : XY pair (tuple, array or list)

        size : w,h pair (tuple, array or list)
            The size for the scale defines the area taken up by the line and
            the ticks.
            This also controls whether the scale is horizontal or vertical.

        units : the units to interpret the pos and size

        flip : bool
            By default the labels will be below or left of the line. This
            puts them above (or right)

        granularity : int or float
            The smallest valid increments for the scale. 0 gives a continuous
            (e.g. "VAS") scale. 1 gives a traditional likert scale. Something
            like 0.1 gives a limited fine-grained scale.

        labelColor / color :
            Color of the labels according to the color space

        markerColor / fillColor :
            Color of the marker according to the color space

        lineColor / borderColor :
            Color of the line and ticks according to the color space

        font : font name

        autodraw :

        depth :

        name :

        autoLog :
        """
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        super(Slider, self).__init__(name=name, autoLog=False)

        self.win = win
        if ticks is None:
            self.ticks = None
        else:
            self.ticks = np.array(ticks)
        self.labels = labels
        # Set pos and size via base method as objects don't yet exist to layout
        self.units = units
        WindowMixin.pos.fset(self, pos)
        if size is None:
            size = layout.Size((1, 0.1), 'height', self.win)
        WindowMixin.size.fset(self, size)
        # Set multiplier and additions to each component's size
        self._markerSizeMultiplier = (1, 1)
        self._markerSizeAddition = (0, 0)
        self._lineSizeMultiplier = (1, 1)
        self._lineSizeAddition = (0, 0)
        self._tickSizeMultiplier = (1, 1)
        self._tickSizeAddition = (0, 0)

        self.granularity = granularity
        self.colorSpace = colorSpace
        self.color = color if color is not False else labelColor
        self.fillColor = fillColor if fillColor is not False else markerColor
        self.borderColor = borderColor if borderColor is not False else lineColor
        self.opacity = opacity
        self.font = font
        self.autoDraw = autoDraw
        self.depth = depth
        self.name = name
        self.autoLog = autoLog
        self.readOnly = readOnly
        self.ori = ori
        self.flip = flip

        self.startValue = self.markerPos = startValue
        self.rt = None
        self.history = []
        self.marker = None
        self.line = None
        self.tickLines = None
        self.labelWrapWidth = labelWrapWidth
        self.labelHeight = labelHeight or min(self.size) / 2
        self._updateMarkerPos = True
        self._dragging = False
        self.mouse = event.Mouse(win=win)
        self._mouseStateClick = None  # so we can rule out long click probs
        self._mouseStateXY = None  # so we can rule out long click probs

        self.validArea = None
        # Create elements
        self._createElements()
        self.styleTweaks = []
        self.style = style
        self.styleTweaks += styleTweaks
        self._layout()
        # some things must wait until elements created
        self.contrast = 1.0

        # set autoLog (now that params have been initialised)
        self.autoLog = autoLog
        if autoLog:
            logging.exp("Created %s = %s" % (self.name, repr(self)))
        self.status = NOT_STARTED
        self.responseClock = core.Clock()

    def __repr__(self, complete=False):
        return self.__str__(complete=complete)  # from MinimalStim

    @property
    def _tickL(self):
        """The length of the line (in the size units)
        """
        return min(self.extent)

    @property
    def units(self):
        return WindowMixin.units.fget(self)

    @units.setter
    def units(self, value):
        WindowMixin.units.fset(self, value)
        if hasattr(self, "line"):
            self.line.units = value
        if hasattr(self, "marker"):
            self.marker.units = value
        if hasattr(self, "tickLines"):
            self.tickLines.units = value
        if hasattr(self, "labelObjs"):
            for label in self.labelObjs:
                label.units = value
        if hasattr(self, "validArea"):
            self.validArea.units = value

    @property
    def pos(self):
        return WindowMixin.pos.fget(self)

    @pos.setter
    def pos(self, value):
        WindowMixin.pos.fset(self, value)
        self._layout()

    @property
    def size(self):
        return WindowMixin.size.fget(self)

    @size.setter
    def size(self, value):
        WindowMixin.size.fset(self, value)
        self._layout()

    @property
    def horiz(self):
        """(readonly) determines from self.size whether the scale is horizontal"""
        return self.extent[0] > self.extent[1]

    @property
    def categorical(self):
        """(readonly) determines from labels and ticks whether the slider is categorical"""
        return self.ticks is None

    @property
    def extent(self):
        """
        The distance from the leftmost point on the slider to the rightmost point, and from the highest
        point to the lowest.
        """
        # Get orientation (theta) and inverse orientation (atheta) in radans
        theta = np.radians(self.ori)
        atheta = np.radians(90-self.ori)
        # Calculate adjacent sides to get vertical extent
        A1 = abs(np.cos(theta) * self.size[1])
        A2 = abs(np.cos(atheta) * self.size[0])
        # Calculate opposite sides to get horizontal extent
        O1 = abs(np.sin(theta) * self.size[1])
        O2 = abs(np.sin(atheta) * self.size[0])
        # Return extent
        return O1 + O2, A1 + A2

    @extent.setter
    def extent(self, value):
        self._extent = layout.Size(self.extent, self.units, self.win)

    @property
    def flip(self):
        if hasattr(self, "_flip"):
            return self._flip

    @flip.setter
    def flip(self, value):
        self._flip = value

    @property
    def opacity(self):
        BaseVisualStim.opacity.fget(self)

    @opacity.setter
    def opacity(self, value):
        BaseVisualStim.opacity.fset(self, value)
        self.fillColor = self._fillColor.copy()
        self.borderColor = self._borderColor.copy()
        self.foreColor = self._foreColor.copy()

    def setOpacity(self, value):
        self.opacity = value

    @property
    def labelHeight(self):
        if hasattr(self, "_labelHeight"):
            return getattr(self._labelHeight, self.units)[1]

    @labelHeight.setter
    def labelHeight(self, value):
        if isinstance(value, layout.Vector):
            # If given a Size, use it
            self._labelHeight = value
        else:
            # Otherwise, convert to a Size object
            self._labelHeight = layout.Size([None, value], units=self.units, win=self.win)

    @property
    def labelWrapWidth(self):
        if hasattr(self, "_labelWrapWidth"):
            return getattr(self._labelWrapWidth, self.units)[0]

    @labelWrapWidth.setter
    def labelWrapWidth(self, value):
        if value is None:
            pass
        elif isinstance(value, layout.Vector):
            # If given a Size, use it
            self._labelWrapWidth = value
        else:
            # Otherwise, convert to a Size object
            self._labelWrapWidth = layout.Size([value, None], units=self.units, win=self.win)

    @property
    def foreColor(self):
        ColorMixin.foreColor.fget(self)

    @foreColor.setter
    def foreColor(self, value):
        ColorMixin.foreColor.fset(self, value)
        # Set color of each label
        if hasattr(self, 'labelObjs'):
            for lbl in self.labelObjs:
                lbl.color = self._foreColor.copy()

    @property
    def labelColor(self):
        """
        Synonym of Slider.foreColor
        """
        return self.foreColor

    @labelColor.setter
    def labelColor(self, value):
        self.foreColor = value

    @property
    def fillColor(self):
        ColorMixin.fillColor.fget(self)

    @fillColor.setter
    def fillColor(self, value):
        ColorMixin.fillColor.fset(self, value)
        # Set color of marker
        if hasattr(self, 'marker'):
            self.marker.fillColor = self._fillColor.copy()

    @property
    def markerColor(self):
        """
        Synonym of Slider.fillColor
        """
        return self.fillColor

    @markerColor.setter
    def markerColor(self, value):
        self.fillColor = value

    @property
    def borderColor(self):
        ColorMixin.borderColor.fget(self)

    @borderColor.setter
    def borderColor(self, value):
        ColorMixin.borderColor.fset(self, value)
        # Set color of lines
        if hasattr(self, 'line'):
            if self.style not in ["scrollbar"]: # Scrollbar doesn't have an outline
                self.line.color = self._borderColor.copy()
            self.line.fillColor = self._borderColor.copy()
            if self.style in ["slider", "scrollbar"]: # Slider and scrollbar need translucent fills
                self.line._fillColor.alpha *= 0.2
        if hasattr(self, 'tickLines'):
            self.tickLines.colors = self._borderColor.copy()
            self.tickLines.opacities = self._borderColor.alpha

    def reset(self):
        """Resets the slider to its starting state (so that it can be restarted
        on each trial with a new stimulus)
        """
        self.markerPos = self.startValue
        self.rating = None  # this is reset to None, whatever the startValue
        self.history = []
        self.rt = None
        self.responseClock.reset()
        self.status = NOT_STARTED

    def _createElements(self):
        # Refresh extent
        self.extent = self.extent

        # Make line
        self._getLineParams()
        self.line = Rect(
            win=self.win,
            pos=self.lineParams['pos'], size=self.lineParams['size'], units=self.units,
            fillColor=self._borderColor.copy(), colorSpace=self.colorSpace,
            autoLog=False
        )
        # Make ticks
        self._getTickParams()
        self.tickLines = ElementArrayStim(
            win=self.win,
            xys=self.tickParams['xys'], sizes=self.tickParams['sizes'], units=self.units,
            nElements=len(self.ticks), elementMask=None, sfs=0,
            colors=self._borderColor.copy(), opacities=self._borderColor.alpha, colorSpace=self.colorSpace,
            autoLog=False
        )
        # Make labels
        self.labelObjs = []
        if self.labels is not None:
            self._getLabelParams()
            for n, label in enumerate(self.labels):
                # Skip blank labels
                if label is None:
                    continue
                # Create textbox for each label
                obj = TextBox2(
                    self.win, label, font=self.font,
                    pos=self.labelParams['pos'][n], size=self.labelParams['size'][n], padding=self.labelParams['padding'][n], units=self.units,
                    anchor=self.labelParams['anchor'][n], alignment=self.labelParams['alignment'][n],
                    color=self._foreColor.copy(), fillColor=None, colorSpace=self.colorSpace,
                    borderColor="red" if debug else None,
                    letterHeight=self.labelHeight,
                    autoLog=False
                )
                self.labelObjs.append(obj)
        # Make marker
        self._getMarkerParams()
        self.marker = ShapeStim(
            self.win,
            vertices="circle",
            pos=self.markerParams['pos'], size=self.markerParams['size'], units=self.units,
            fillColor=self._fillColor, lineColor=None,
            autoLog=False
        )

        # create a rectangle to check for clicks
        self._getHitboxParams()
        self.validArea = Rect(
            self.win,
            pos=self.hitboxParams['pos'], size=self.hitboxParams['size'], units=self.units,
            fillColor=None, lineColor="red" if debug else None,
            autoLog=False
        )

    def _layout(self):
        # Refresh style
        self.style = self.style
        # Refresh extent
        self.extent = self.extent

        # Get marker params
        self._getMarkerParams()
        # Apply marker params
        self.marker.units = self.units
        for param, value in self.markerParams.items():
            setattr(self.marker, param, value)

        # Get line params
        self._getLineParams()
        # Apply line params
        self.line.units = self.units
        for param, value in self.lineParams.items():
            setattr(self.line, param, value)

        # Get tick params
        self._getTickParams()
        # Apply tick params
        self.tickLines.units = self.units
        for param, value in self.tickParams.items():
            setattr(self.tickLines, param, value)

        # Get label params
        self._getLabelParams()
        # Apply label params
        for n, obj in enumerate(self.labelObjs):
            obj.units = self.units
            for param, value in self.labelParams.items():
                setattr(obj, param, value[n])

        # Get hitbox params
        self._getHitboxParams()
        # Apply hitbox params
        self.validArea.units = self.units
        for param, value in self.hitboxParams.items():
            setattr(self.validArea, param, value)

    def _ratingToPos(self, rating):
        # Get ticks or substitute
        if self.ticks is not None:
            ticks = self.ticks
        else:
            ticks = [0, 1]
        # Reshape rating to handle multiple values
        rating = np.array(rating)
        rating = rating.reshape((-1, 1))
        rating = np.hstack((rating, rating))
        # Adjust to scale bottom
        magDelta = rating - ticks[0]
        # Adjust to scale magnitude
        delta = magDelta / (ticks[-1] - ticks[0])
        # Adjust to scale size
        delta = self._extent.pix * (delta - 0.5)
        # Adjust to scale pos
        pos = delta + self._pos.pix
        # Replace irrelevant value according to orientation
        pos[:, int(self.horiz)] = self._pos.pix[int(self.horiz)]
        # Convert to native units
        return getattr(layout.Position(pos, units="pix", win=self.win), self.units)

    def _posToRating(self, pos):
        # Get ticks or substitute
        if self.ticks is not None:
            ticks = self.ticks
        else:
            ticks = [0, 1]
        # Get in pix
        pos = layout.Position(pos, units=self.win.units, win=self.win).pix
        # Get difference from scale pos
        delta = pos - self._pos.pix
        # Adjust to scale size
        delta = delta / self._extent.pix + 0.5
        # Adjust to scale magnitude
        magDelta = delta * (ticks[-1] - ticks[0])
        # Adjust to scale bottom
        rating = magDelta + ticks[0]
        # Return relevant value according to orientation
        return rating[1-int(self.horiz)]

    def _getLineParams(self):
        """
        Calculates location and size of the line based on own location and size
        """
        # Store line details
        self.lineParams = {
            'units': self.units,
            'pos': self.pos,
            'size': self._extent * np.array(self._lineSizeMultiplier) + layout.Size(self._lineSizeAddition, self.units, self.win)
        }

    def _getMarkerParams(self):
        """
        Calculates location and size of marker based on own location and size
        """
        # Calculate pos
        pos = self._ratingToPos(self.rating or 0)
        # Get size (and correct for norm)
        size = layout.Size([min(self._size.pix)]*2, 'pix', self.win)
        # Store marker details
        self.markerParams = {
            'units': self.units,
            'pos': pos,
            'size': size * np.array(self._markerSizeMultiplier) + layout.Size(self._markerSizeAddition, self.units, self.win),
        }

    def _getTickParams(self):
        """ Calculates the locations of the line, tickLines and labels from
        the rating info
        """
        # If categorical, create tick values from labels
        if self.categorical:
            self.ticks = np.arange(len(self.labels))
            self.granularity = 1.0
        # Calculate positions
        xys = self._ratingToPos(self.ticks)
        # Get size (and correct for norm)
        size = layout.Size([min(self._extent.pix)]*2, 'pix', self.win)
        # Store tick details
        self.tickParams = {
            'units': self.units,
            'xys': xys,
            'sizes': np.tile(
                getattr(size, self.units) * np.array(self._tickSizeMultiplier) + np.array(self._tickSizeAddition),
                (len(self.ticks), 1)),
        }

    def _getLabelParams(self):
        if self.labels is None:
            return
        # Get number of labels now for convenience
        n = len(self.labels)
        # Get coords of slider edges
        top = self.pos[1] + self.extent[1] / 2
        bottom = self.pos[1] - self.extent[1] / 2
        left = self.pos[0] - self.extent[0] / 2
        right = self.pos[0] + self.extent[0] / 2
        # Work out where labels are relative to line
        w = self.labelWrapWidth
        if self.horiz:  # horizontal
            # Always centered horizontally
            anchorHoriz = alignHoriz = 'center'
            # Width as fraction of size, height starts at double slider
            if w is None:
                w = self.extent[0] / len(self.ticks)
            h = self.extent[1] * 3
            # Evenly spaced, constant y
            x = np.linspace(left, right, num=n)
            x = arraytools.snapto(x, points=self.tickParams['xys'][:, 0])
            y = [self.pos[1]] * n
            # Padding applied on vertical
            paddingHoriz = 0
            paddingVert = (self._tickL + self.labelHeight) / 2
            # Vertical align/anchor depend on flip
            if not self.flip:
                # Labels below means anchor them from the top
                anchorVert = alignVert = 'top'
            else:
                # Labels on top means anchor them from below
                anchorVert = alignVert = 'bottom'
        else:  # vertical
            # Always centered vertically
            anchorVert = alignVert = 'center'
            # Height as fraction of size, width starts at double slider
            h = self.extent[1] / len(self.ticks)
            if w is None:
                w = self.extent[0] * 3
            # Evenly spaced and clipped to ticks, constant x
            y = np.linspace(bottom, top, num=n)
            y = arraytools.snapto(y, points=self.tickParams['xys'][:, 1])
            x = [self.pos[0]] * n
            # Padding applied on horizontal
            paddingHoriz = (self._tickL + self.labelHeight) / 2
            paddingVert = 0
            # Horizontal align/anchor depend on flip
            if not self.flip:
                # Labels left means anchor them from the right
                anchorHoriz = alignHoriz = 'right'
            else:
                # Labels right means anchor them from the left
                anchorHoriz = alignHoriz = 'left'
        # Store label details
        self.labelParams = {
            'units': (self.units,) * n,
            'pos': np.vstack((x, y)).transpose(None),
            'size': np.tile((w, h), (n, 1)),
            'padding': np.tile((paddingHoriz, paddingVert), (n, 1)),
            'anchor': np.tile((anchorHoriz, anchorVert), (n, 1)),
            'alignment': np.tile((alignHoriz, alignVert), (n, 1))
        }

    def _getHitboxParams(self):
        """
        Calculates hitbox size and pos from own size and pos
        """
        # Get pos
        pos = self.pos
        # Get size
        size = self._extent * 1.1
        # Store hitbox details
        self.hitboxParams = {
            'units': self.units,
            'pos': pos,
            'size': size,
        }

    def _granularRating(self, rating):
        """Handle granularity for the rating"""
        if rating is not None:
            if self.granularity > 0:
                rating = round(rating / self.granularity) * self.granularity
                rating = round(rating, 8)  # or gives 1.9000000000000001
            rating = max(rating, self.ticks[0])
            rating = min(rating, self.ticks[-1])
        return rating

    @property
    def rating(self):
        if hasattr(self, "_rating"):
            return self._rating

    @rating.setter
    def rating(self, rating):
        """The most recent rating from the participant or None.
        Note that the position of the marker can be set using current without
        looking like a change in the marker position"""
        rating = self._granularRating(rating)
        self.markerPos = rating
        if self.categorical and (rating is not None):
            rating = self.labels[int(round(rating))]
        self._rating = rating

    @property
    def value(self):
        """Synonymous with .rating"""
        return self.rating
    @value.setter
    def value(self, val):
        self.rating = val

    @attributeSetter
    def markerPos(self, rating):
        """The position on the scale where the marker should be. Note that
        this does not alter the value of the reported rating, only its visible
        display.
        Also note that this position is in scale units, not in coordinates"""
        rating = self._granularRating(rating)
        if ('markerPos' not in self.__dict__ or not np.alltrue(
                self.__dict__['markerPos'] == rating)):
            self.__dict__['markerPos'] = rating
            self._updateMarkerPos = True

    def recordRating(self, rating, rt=None, log=None):
        """Sets the current rating value
        """
        rating = self._granularRating(rating)
        setAttribute(self, attrib='rating', value=rating, operation='', log=log)
        if rt is None:
            self.rt = self.responseClock.getTime()
        else:
            self.rt = rt
        self.history.append((rating, self.rt))
        self._updateMarkerPos = True

    def getRating(self):
        """Get the current value of rating (or None if no response yet)
        """
        return self.rating

    def getRT(self):
        """Get the RT for most recent rating (or None if no response yet)
        """
        return self.rt

    def getMarkerPos(self):
        """Get the current marker position (or None if no response yet)
        """
        return self.markerPos

    def setMarkerPos(self, rating):
        """Set the current marker position (or None if no response yet)

        Parameters
        ----------
        rating : int or float
            The rating on the scale where we want to set the marker
        """
        if self._updateMarkerPos:
            self.marker.pos = self._ratingToPos(rating)
            self.markerPos = rating
            self._updateMarkerPos = False

    def draw(self):
        """Draw the Slider, with all its constituent elements on this frame
        """
        self.getMouseResponses()
        if debug:
            self.validArea.draw()
        self.line.draw()
        self.tickLines.draw()
        if self.markerPos is not None:
            if self._updateMarkerPos:
                self.marker.pos = self._ratingToPos(self.markerPos)
                self._updateMarkerPos = False
            self.marker.draw()
        for label in self.labelObjs:
            label.draw()
        # we started drawing to reset clock on flip
        if self.status == NOT_STARTED:
            self.win.callOnFlip(self.responseClock.reset)
            self.status = STARTED

    def getHistory(self):
        """Return a list of the subject's history as (rating, time) tuples.

        The history can be retrieved at any time, allowing for continuous
        ratings to be obtained in real-time. Both numerical and categorical
        choices are stored automatically in the history.
        """
        return self.history

    def setReadOnly(self, value=True, log=None):
        """When the rating scale is read only no responses can be made and the
        scale contrast is reduced

        Parameters
        ----------
        value : bool (True)
            The value to which we should set the readOnly flag
        log : bool or None
            Force the autologging to occur or leave as default

        """
        setAttribute(self, 'readOnly', value, log)
        if value == True:
            self.contrast = 0.5
        else:
            self.contrast = 1.0

    @attributeSetter
    def contrast(self, contrast):
        """Set all elements of the Slider (labels, ticks, line) to a contrast

        Parameters
        ----------
        contrast
        """
        self.marker.contrast = contrast
        self.line.contrast = contrast
        self.tickLines.contrasts = contrast
        for label in self.labelObjs:
            label.contrast = contrast

    def getMouseResponses(self):
        """Instructs the rating scale to check for valid mouse responses.

        This is usually done during the draw() method but can be done by the
        user as well at any point in time. The rating will be returned but
        will ALSO automatically be set as the current rating response.

        While the mouse button is down we will alter self.markerPos
        but don't set a value for self.rating until button comes up

        Returns
        ----------
        A rating value or None
        """
        if self.readOnly:
            return
        click = bool(self.mouse.getPressed()[0])
        xy = self.mouse.getPos()

        if click:
            # Update current but don't set Rating (mouse is still down)
            # Dragging has to start inside a "valid" area (i.e., on the
            # slider), but may continue even if the mouse moves away from
            # the slider, as long as the mouse button is not released.
            if (self.validArea.contains(self.mouse, units=self.units) or
                    self._dragging):
                self.markerPos = self._posToRating(xy)  # updates marker
                self._dragging = True
            self._updateMarkerPos = True
        else:  # mouse is up - check if it *just* came up
            if self._dragging:
                self._dragging = False
                if self.markerPos is not None:
                    self.recordRating(self.markerPos)
                return self.markerPos
            else:
                # is up and was already up - move along
                return None

        self._mouseStateXY = xy

    # Overload color setters so they set sub-components
    @property
    def foreColor(self):
        ColorMixin.foreColor.fget(self)
    @foreColor.setter
    def foreColor(self, value):
        ColorMixin.foreColor.fset(self, value)
        # Set color for all labels
        if hasattr(self, "labelObjs"):
            for obj in self.labelObjs:
                obj.color = self._foreColor.copy()

    @property
    def fillColor(self):
        ColorMixin.fillColor.fget(self)
    @fillColor.setter
    def fillColor(self, value):
        ColorMixin.fillColor.fset(self, value)
        # Set color for marker
        if hasattr(self, "marker"):
            self.marker.fillColor = self._fillColor.copy()

    @property
    def borderColor(self):
        ColorMixin.borderColor.fget(self)
    @borderColor.setter
    def borderColor(self, value):
        ColorMixin.borderColor.fset(self, value)
        # Set color for lines
        if hasattr(self, "line"):
            self.line.color = self._borderColor.copy()
        if hasattr(self, "tickLines"):
            self.tickLines.colors = self._borderColor.copy()

    knownStyles = ['slider', 'rating', 'radio', 'scrollbar']
    legacyStyles = []
    knownStyleTweaks = ['labels45', 'triangleMarker']
    legacyStyleTweaks = ['whiteOnBlack']

    @property
    def style(self):
        if hasattr(self, "_style"):
            return self._style

    @style.setter
    def style(self, style):
        """Sets some predefined styles or use these to create your own.

        If you fancy creating and including your own styles that would be great!

        Parameters
        ----------
        style: string

            Known styles currently include:

                'rating': the marker is a circle
                'slider': looks more like an application slider control
                'whiteOnBlack': a sort of color-inverse rating scale
                'scrollbar': looks like a scrollbar for a window

            Styles cannot be combined in a list - they are discrete

        """
        self._style = style

        # Legacy: If given a list (as was once the case), take the first style
        if isinstance(style, (list, tuple)):
            styles = style
            style = "rating"
            for val in styles:
                # If list contains a style, use it
                if val in self.knownStyles + self.legacyStyles:
                    style = val
                # Apply any tweaks
                if val in self.knownStyleTweaks + self.legacyStyleTweaks:
                    self.styleTweaks += val

        if style == 'rating' or style is None:
            # Narrow line
            self.line.opacity = 1
            self._lineSizeAddition = (0, 0)
            if self.horiz:
                self._lineSizeMultiplier = (1, 0.1)
            else:
                self._lineSizeMultiplier = (0.1, 1)
            # 1:1 circular markers
            self.marker.vertices = "circle"
            self._markerSizeMultiplier = (1, 1)
            self._markerSizeAddition = (0, 0)
            # Narrow rectangular ticks
            self.tickLines.elementMask = None
            self._tickSizeAddition = (0, 0)
            if self.horiz:
                self._tickSizeMultiplier = (0.1, 1)
            else:
                self._tickSizeMultiplier = (1, 0.1)

        if style == 'slider':
            # Semi-transparent rectangle for a line
            self.line._fillColor.alpha = 0.2
            self._lineSizeMultiplier = (1, 1)
            self._lineSizeAddition = (0, 0)
            # Rectangular marker
            self.marker.vertices = "rectangle"
            self._markerSizeAddition = (0, 0)
            if self.horiz:
                self._markerSizeMultiplier = (0.1, 0.8)
            else:
                self._markerSizeMultiplier = (0.8, 0.1)
            # Narrow rectangular ticks
            self.tickLines.elementMask = None
            self._tickSizeAddition = (0, 0)
            if self.horiz:
                self._tickSizeMultiplier = (0.1, 1)
            else:
                self._tickSizeMultiplier = (1, 0.1)

        if style == 'radio':
            # No line
            self._lineSizeMultiplier = (0, 0)
            self._lineSizeAddition = (0, 0)
            # 0.7 scale circular markers
            self.marker.vertices = "circle"
            self._markerSizeMultiplier = (0.7, 0.7)
            self._markerSizeAddition = (0, 0)
            # 1:1 circular ticks
            self.tickLines.elementMask = 'circle'
            self._tickSizeMultiplier = (1, 1)
            self._tickSizeAddition = (0, 0)

        if style == 'scrollbar':
            # Semi-transparent rectangle for a line (+ extra area for marker)
            self.line.opacity = 1
            self.line._fillColor.alpha = 0.2
            self._lineSizeAddition = (0, 0)
            if self.horiz:
                self._lineSizeMultiplier = (1.2, 1)
            else:
                self._lineSizeMultiplier = (1, 1.2)
            # Long rectangular marker
            self.marker.vertices = "rectangle"
            if self.horiz:
                self._markerSizeMultiplier = (0, 1)
                self._markerSizeAddition = (self.extent[0] / 5, 0)
            else:
                self._markerSizeMultiplier = (1, 0)
                self._markerSizeAddition = (0, self.extent[1] / 5)
            # No ticks
            self.tickLines.elementMask = None
            self._tickSizeAddition = (0, 0)
            self._tickSizeMultiplier = (0, 0)

        # Legacy: If given a tweak, apply it as a tweak rather than a style
        if style in self.knownStyleTweaks + self.legacyStyleTweaks:
            self.styleTweaks.append(style)

        # Refresh style tweaks (as these override some aspects of style)
        self.styleTweaks = self.styleTweaks

        return style

    @attributeSetter
    def styleTweaks(self, styleTweaks):
        """Sets some predefined style tweaks or use these to create your own.

        If you fancy creating and including your own style tweaks that would be great!

        Parameters
        ----------
        styleTweaks: list of strings

            Known style tweaks currently include:

                'triangleMarker': the marker is a triangle
                'labels45': the text is rotated by 45 degrees

            Legacy style tweaks include:

                'whiteOnBlack': a sort of color-inverse rating scale

            Legacy style tweaks will work if set in code, but are not exposed in Builder as they are redundant

            Style tweaks can be combined in a list e.g. `['labels45']`

        """
        if not isinstance(styleTweaks, (list, tuple, np.ndarray)):
            styleTweaks = [styleTweaks]

        self.__dict__['styleTweaks'] = styleTweaks

        if 'triangleMarker' in styleTweaks:
            if self.horiz and self.flip:
                ori = -90
            elif self.horiz:
                ori = -90
            elif not self.horiz and self.flip:
                ori = 180
            else:
                ori = 0

            self.marker.vertices = [[0, 0], [0.5, 0.5], [0.5, -0.5]]

        if 'labels45' in styleTweaks:
            for label in self.labelObjs:
                label.ori = -45

        # Legacy
        if 'whiteOnBlack' in styleTweaks:
            self.line.color = 'black'
            self.tickLines.colors = 'black'
            self.marker.color = 'white'
            for label in self.labelObjs:
                label.color = 'white'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Determine screen gamma using motion-nulling method
of Ledgeway and Smith, 1994, Vision Research, 34, 2727-2740
A similar system had been used early for chromatic isoluminance:
Anstis SM, Cavanagh P. A minimum motion technique for judging equiluminance.
In: Sharpe MJD & LT Colour vision: Psychophysics and physiology. London: Academic Press; 1983. pp. 66-77.

Instructions: on each trial press the up/down cursor keys depending on
the apparent direction of motion of the bars.
"""

from psychopy import visual, core, event, gui, data
from psychopy.tools.filetools import fromFile, toFile
from psychopy.visual import filters
import numpy as num
import time

try:
    # try to load previous info
    info = fromFile('info_gamma.pickle')
    print(info)
except Exception:
    # if no file use some defaults
    info = {}
    info['lumModNoise'] = 0.5
    info['lumModLum'] = 0.1
    info['contrastModNoise'] = 1.0
    info['observer'] = ''
    info['highGamma'] = 3.0
    info['lowGamma'] = 0.8
    info['nTrials'] = 50
dlg = gui.DlgFromDict(info)
# save to a file for future use (ie storing as defaults)
if dlg.OK:
    toFile('info_gamma.pickle', info)
else:
    core.quit()  # user cancelled. quit
print(info)

info['timeStr']=time.strftime("%b_%d_%H%M", time.localtime())
nFrames = 3
cyclesTime = 2
cyclesSpace = 2
pixels = 128

win = visual.Window((1024, 768), units='pix', allowGUI=True, bitsMode=None)
visual.TextStim(win, text='building stimuli').draw()

win.flip()

globalClock = core.Clock()

# for luminance modulated noise
noiseMatrix = num.random.randint(0, 2, [pixels, pixels])  # * noiseContrast
noiseMatrix = noiseMatrix * 2.0-1  # into range -1: 1

stimFrames = []; lumGratings=[]
# create the 4 frames of the sequence (luminance and contrast modulated noise in quadrature)
lumGratings.append(filters.makeGrating(pixels, 0, cyclesSpace, phase=0))
stimFrames.append(visual.GratingStim(win, texRes=pixels, mask='circle',
        size=pixels * 2, sf=1.0 / pixels, ori=90,
        tex= (noiseMatrix * info['lumModNoise'] + lumGratings[0] * info['lumModLum'])
        ))
lumGratings.append(filters.makeGrating(pixels, 0, cyclesSpace, phase=90) / 2.0 + 0.5)
stimFrames.append(visual.GratingStim(win, texRes=pixels, mask='circle',
        size=pixels * 2, sf=1.0/pixels, ori=90,
        tex= (noiseMatrix * info['contrastModNoise'] * lumGratings[1])
        ))
lumGratings.append(filters.makeGrating(pixels, 0, cyclesSpace, phase=180))
stimFrames.append(visual.GratingStim(win, texRes=pixels, mask='circle',
        size=pixels * 2, sf=1.0/pixels, ori=90,
        tex= (noiseMatrix * info['lumModNoise'] + lumGratings[2] * info['lumModLum'])
        ))
lumGratings.append(filters.makeGrating(pixels, 0, cyclesSpace, phase=270) / 2.0 + 0.5)
stimFrames.append(visual.GratingStim(win, texRes=pixels, mask='circle',
        size=pixels * 2, sf=1.0/pixels, ori=90,
        tex= (noiseMatrix * info['contrastModNoise'] * lumGratings[3])
        ))

stairCases = []
# two staircases - one from the top, one from below - to average
stairCases.append(data.StairHandler(startVal=info['highGamma'], nTrials=info['nTrials'],
        stepSizes=[0.5, 0.5, 0.1, 0.1, 0.1, 0.1, 0.05, 0.05], stepType='lin',
        nUp=1, nDown=1))
stairCases.append(data.StairHandler(startVal=info['lowGamma'], nTrials=info['nTrials'],
        stepSizes=[0.5, 0.5, 0.1, 0.1, 0.1, 0.1, 0.05, 0.05], stepType='lin',
        nUp=1, nDown=1))

def getResponse(direction):
    """if subject said up when direction was up ( + 1) then increase gamma
    Otherwise, decrease gamma"""
    event.clearEvents()  # clear the event buffer to start with

    while 1:  # forever until we return
        for key in event.getKeys():
            # quit
            if key in ['escape', 'q']:
                win.close()
                # win.bits.reset()
                core.quit()
            # valid response - check to see if correct
            elif key in ['down', 'up']:
                if ((key in ['down'] and direction == -1) or
                        (key in ['up'] and direction == +1)):
                    return 0
                else:
                    return 1
            else:
                print("hit DOWN or UP (or Esc) (You hit %s)" %key)

def presentStimulus(direction):
    """Present stimulus drifting in a given direction (for low gamma)
    where:
        direction = + 1(up) or -1(down)
    """
    win.fps()

    startPhase = num.random.random()
    if direction == 1:
        frameIndices = num.arange(0, 4)
    else:
        frameIndices = num.arange(3, -1, -1)

    for cycles in range(cyclesTime):
        # cycle through the 4 frames
        for ii in frameIndices:
            thisStim = stimFrames[ii]
            thisStim.setPhase(startPhase)
            for n in range(nFrames):
                # present for several constant frames (TF)
                thisStim.draw()
                win.flip()

    # then blank the screen
    win.flip()

# run the staircase
for trialN in range(info['nTrials']):
    for stairCase in stairCases:
        thisGamma = next(stairCase)
        t = globalClock.getTime()
        win.gamma = [thisGamma, thisGamma, thisGamma]

        direction = num.random.randint(0, 2) * 2-1  # a random number -1 or 1
        presentStimulus(direction)

        ans = getResponse(direction)
        stairCase.addData(ans)

win.flip()
core.wait(0.5)

win.close()

# save data
fileName = gui.fileSaveDlg('.', '%s_%s' %(info['observer'], info['timeStr']))
stairCases[1].saveAsPickle(fileName + 'hi')
stairCases[1].saveAsText(fileName + 'hi')
stairCases[0].saveAsPickle(fileName + 'lo')
stairCases[0].saveAsText(fileName + 'lo')

print('That took %.1fmins' % (globalClock.getTime() / 60.0))

core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo: Rotate flashing wedge
"""

from psychopy import visual, event, core

win = visual.Window([800, 800])
globalClock = core.Clock()

# Make two wedges (in opposite contrast) and alternate them for flashing
wedge1 = visual.RadialStim(win, tex='sqrXsqr', color=1, size=1,
    visibleWedge=[0, 45], radialCycles=4, angularCycles=8, interpolate=False,
    autoLog=False)  # this stim changes too much for autologging to be useful
wedge2 = visual.RadialStim(win, tex='sqrXsqr', color=-1, size=1,
    visibleWedge=[0, 45], radialCycles=4, angularCycles=8, interpolate=False,
    autoLog=False)  # this stim changes too much for autologging to be useful

t = 0
rotationRate = 0.1  # revs per sec
flashPeriod = 0.1  # seconds for one B-W cycle (ie 1/Hz)
while not event.getKeys():
    t = globalClock.getTime()
    if t % flashPeriod < flashPeriod / 2.0:  # more accurate to count frames
        stim = wedge1
    else:
        stim = wedge2
    stim.ori = t * rotationRate * 360.0  # set new rotation
    stim.draw()
    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



from pathlib import Path
from tempfile import mkdtemp

from psychopy import experiment
from . import _TestDisabledMixin, _TestBaseComponentsMixin
from psychopy.experiment.loops import TrialHandler
from psychopy.experiment.routines import Routine
from psychopy.experiment.components.code import CodeComponent
from psychopy.tests.utils import TESTS_DATA_PATH


class TestCodeComponent(_TestBaseComponentsMixin, _TestDisabledMixin):
    """
    Test that Code coponents have the correct params and write as expected.
    """

    @classmethod
    def setup_class(cls):
        cls.exp = experiment.Experiment() # create once, not every test
        try:
            cls.tempDir = mkdtemp(dir=Path(__file__).root, prefix='psychopy-tests-app')
        except (PermissionError, OSError):
            # can't write to root on Linux
            cls.tempDir = mkdtemp(prefix='psychopy-tests-app')

    def setup(self):
        # Make blank experiment
        self.exp = experiment.Experiment()
        # Make blank routine
        self.routine = Routine(name="testRoutine", exp=self.exp)
        self.exp.addRoutine("testRoutine", self.routine)
        self.exp.flow.addRoutine(self.routine, 0)
        # Add loop around routine
        self.loop = TrialHandler(exp=self.exp, name="testLoop")
        self.exp.flow.addLoop(self.loop, 0, -1)
        # Make Mouse component
        self.comp = CodeComponent(exp=self.exp, parentName="testRoutine", name="testCode")
        self.routine.addComponent(self.comp)

    def test_all_code_component_tabs(self):
        # Names of each tab in a Code component
        tabs = {
            'Before Experiment': '___before_experiment___',
            'Begin Experiment': '___begin_experiment___',
            'Begin Routine': '___begin_routine___',
            'Each Frame': '___each_frame___',
            'End Routine': '___end_routine___',
            'End Experiment': '___end_experiment___',
        }
        # Add markers to component
        for paramName, marker in tabs.items():
            jsParamName = paramName.replace(" ", " JS ")
            self.comp.params[paramName].val = self.comp.params[jsParamName].val = marker

        # Write script
        pyScript = self.exp.writeScript(target="PsychoPy")
        jsScript = self.exp.writeScript(target="PsychoJS")

        # Check that code from each tab exists in compiled script
        for lang, script in {"Python": pyScript, "JS": jsScript}.items():
            for paramName, marker in tabs.items():
                try:
                    assert marker in script, (
                        f"Could not find {marker} in {lang} script."
                    )
                except AssertionError as err:
                    # If test fails here, save the file for easy access
                    ext = ".py" if lang == "Python" else ".js"
                    with open(Path(TESTS_DATA_PATH) / ("test_all_code_component_tabs_local" + ext), "w") as f:
                        f.write(script)
                    raise err
            if lang == "Python":
                # Check py code is in the right order in Python (not applicable to JS as it's non-linear)
                assert script.find('___before_experiment___') < script.find('___begin_experiment___') < script.find(
                    '___begin_routine___') < script.find('___each_frame___') < script.find('___end_routine___') < script.find(
                    '___end_experiment___')
                assert script.find('___before_experiment___') < script.find('visual.Window') < script.find(
                    '___begin_experiment___') < script.find('continueRoutine = True')
                assert script.find('continueRoutine = True') < script.find('___begin_routine___') < script.find(
                    'while continueRoutine:') < script.find('___each_frame___')
                assert script.find('thisComponent.setAutoDraw(False)') < script.find('___end_routine___') < script.find(
                    'routineTimer.reset()') < script.find('___end_experiment___')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provides read / write access to the parallel port for
Linux or Windows.

The :class:`~psychopy.parallel.Parallel` class described below will
attempt to load whichever parallel port driver is first found on your
system and should suffice in most instances. If you need to use a specific
driver then, instead of using :class:`~psychopy.parallel.ParallelPort`
shown below you can use one of the following as drop-in replacements,
forcing the use of a specific driver:

    - `psychopy.parallel.PParallelInpOut`
    - `psychopy.parallel.PParallelDLPortIO`
    - `psychopy.parallel.PParallelLinux`

Either way, each instance of the class can provide access to a different
parallel port.

There is also a legacy API which consists of the routines which are directly
in this module. That API assumes you only ever want to use a single
parallel port at once.
"""
import sys
from psychopy import logging

# To make life easier, only try drivers which have a hope in heck of working.
# Because hasattr() in connection to windll ends up in an OSError trying to
# load 32bit drivers in a 64bit environment, different drivers defined in
# the dictionary 'drivers' are tested.

if sys.platform.startswith('linux'):
    from ._linux import PParallelLinux
    ParallelPort = PParallelLinux
elif sys.platform == 'win32':
    drivers = dict(inpout32=('_inpout', 'PParallelInpOut'),
                   inpoutx64=('_inpout', 'PParallelInpOut'),
                   dlportio=('_dlportio', 'PParallelDLPortIO'))
    from ctypes import windll
    from importlib import import_module
    for key, val in drivers.items():
        driver_name, class_name = val
        try:
            hasattr(windll, key)
            ParallelPort = getattr(import_module('.'+driver_name, __name__),
                                   class_name)
            break
        except (OSError, KeyError, NameError):
            ParallelPort = None
            continue
    if ParallelPort is None:
        logging.warning("psychopy.parallel has been imported but no "
                        "parallel port driver found. Install either "
                        "inpout32, inpoutx64 or dlportio")
else:
    logging.warning("psychopy.parallel has been imported on a Mac "
                    "(which doesn't have a parallel port?)")

    # macOS doesn't have a parallel port but write the class for doc purps
    class ParallelPort:
        """Class for read/write access to the parallel port on Windows & Linux

        Usage::

            from psychopy import parallel
            port = parallel.ParallelPort(address=0x0378)

            port.setData(4)
            port.readPin(2)
            port.setPin(2, 1)
        """

        def __init__(self, address):
            """This is just a dummy constructor to avoid errors
            when the parallel port cannot be initiated
            """
            msg = ("psychopy.parallel has been imported but (1) no parallel "
                   "port driver could be found or accessed on Windows or "
                   "(2) PsychoPy is run on a Mac (without parallel-port "
                   "support for now)")
            logging.warning(msg)

        def setData(self, data):
            """Set the data to be presented on the parallel port (one ubyte).
            Alternatively you can set the value of each pin (data pins are
            pins 2-9 inclusive) using :func:`~psychopy.parallel.setPin`

            Examples::

                from psychopy import parallel
                port = parallel.ParallelPort(address=0x0378)

                port.setData(0)  # sets all pins low
                port.setData(255)  # sets all pins high
                port.setData(2)  # sets just pin 3 high (pin2 = bit0)
                port.setData(3)  # sets just pins 2 and 3 high

            You can also convert base 2 to int easily in python::

                port.setData( int("00000011", 2) )  # pins 2 and 3 high
                port.setData( int("00000101", 2) )  # pins 2 and 4 high
            """
            sys.stdout.flush()
            raise NotImplementedError("Parallel ports don't work on a Mac")

        def readData(self):
            """Return the value currently set on the data pins (2-9)
            """
            raise NotImplementedError("Parallel ports don't work on a Mac")

        def readPin(self, pinNumber):
            """Determine whether a desired (input) pin is high(1) or low(0).

            Pins 2-13 and 15 are currently read here
            """
            raise NotImplementedError("Parallel ports don't work on a Mac")

# In order to maintain API compatibility, we have to manage
# the old, non-object-based, calls.  This necessitates keeping a
# global object referring to a port.  We initialise it the first time
# that the person calls
PORT = None  # don't create a port until necessary


def setPortAddress(address=0x0378):
    """Set the memory address or device node for your parallel port
    of your parallel port, to be used in subsequent commands

    Common port addresses::

        LPT1 = 0x0378 or 0x03BC
        LPT2 = 0x0278 or 0x0378
        LPT3 = 0x0278

    or for Linux::
        /dev/parport0

    This routine will attempt to find a usable driver depending
    on your platform
    """

    global PORT
    # convert u"0x0378" into 0x0378
    if isinstance(address, str) and address.startswith('0x'):
        address = int(address, 16)

    # This is useful with the Linux-based driver where deleting
    # the port object ensures that we're not longer holding the
    # device node open and that we won't error if we end up
    # re-opening it
    if PORT is not None:
        del PORT

    try:
        PORT = ParallelPort(address=address)
    except Exception as exp:
        logging.warning('Could not initiate port: %s' % str(exp))
        PORT = None


def setData(data):
    """Set the data to be presented on the parallel port (one ubyte).
    Alternatively you can set the value of each pin (data pins are pins
    2-9 inclusive) using :func:`~psychopy.parallel.setPin`

    Examples::

        parallel.setData(0)  # sets all pins low
        parallel.setData(255)  # sets all pins high
        parallel.setData(2)  # sets just pin 3 high (remember that pin2=bit0)
        parallel.setData(3)  # sets just pins 2 and 3 high

    You can also convert base 2 to int v easily in python::

        parallel.setData(int("00000011", 2))  # pins 2 and 3 high
        parallel.setData(int("00000101", 2))  # pins 2 and 4 high
    """

    global PORT
    if PORT is None:
        raise RuntimeError('Port address must be set using setPortAddress')
    PORT.setData(data)


def setPin(pinNumber, state):
    """Set a desired pin to be high (1) or low (0).

    Only pins 2-9 (incl) are normally used for data output::

        parallel.setPin(3, 1)  # sets pin 3 high
        parallel.setPin(3, 0)  # sets pin 3 low
    """
    global PORT
    PORT.setPin(pinNumber, state)


def readPin(pinNumber):
    """Determine whether a desired (input) pin is high(1) or low(0).

    Pins 2-13 and 15 are currently read here
    """
    global PORT
    return PORT.readPin(pinNumber)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""locale handling for PsychoPy experiment scripts, currently Mac 10.10.3+

Purpose: Avoid a unicode-related python crash on Mac 10.10.3 (maybe only
in conda environment?)

Usage: Just import this module at the top of experiment scripts. Should be fast
enough and safe to do for all Builder scripts.

Its unlikely to be widely useful for localizing experiments; that is not its
intended purpose. Few people will want to have the visual display of text in
their experiment vary by locale. If they do, it is easy enough for them to
create multiple versions of an experiment.
"""

import platform
macVer = platform.mac_ver()[0]  # e.g., '10.9.5' or '' for non-Mac

if macVer:
    def _versionTuple(v):
        return tuple(map(int, v.split('.')))
    ver = _versionTuple(macVer)
    if ver > _versionTuple('10.10.2'):
        # set locale and prefs experiment-wide, without saving prefs to disk
        import locale
        from psychopy import prefs
        if not prefs.app['locale']:
            prefs.app['locale'] = u'en_US'
        locale.setlocale(locale.LC_ALL, str(prefs.app['locale']) + '.UTF-8')



from pathlib import Path

from . import _TestDisabledMixin, _TestBaseComponentsMixin
from psychopy.experiment import Experiment
from psychopy.experiment.loops import TrialHandler
from psychopy.experiment.routines import Routine
from psychopy.experiment.components.mouse import MouseComponent
from psychopy.experiment.components.polygon import PolygonComponent
from psychopy.tests.utils import TESTS_DATA_PATH


class TestMouseComponent(_TestBaseComponentsMixin, _TestDisabledMixin):
    """
    Test that Mouse coponents have the correct params and write as expected.
    """
    def setup(self):
        # Make blank experiment
        self.exp = Experiment()
        # Make blank routine
        self.routine = Routine(name="testRoutine", exp=self.exp)
        self.exp.addRoutine("testRoutine", self.routine)
        self.exp.flow.addRoutine(self.routine, 0)
        # Add loop around routine
        self.loop = TrialHandler(exp=self.exp, name="testLoop")
        self.exp.flow.addLoop(self.loop, 0, -1)
        # Make Mouse component
        self.comp = MouseComponent(exp=self.exp, parentName="testRoutine", name="testMouse")
        self.routine.addComponent(self.comp)
        # Make a rect for when we need something to click on
        self.target = PolygonComponent(exp=self.exp, parentName="testRoutine", name="testPolygon")
        self.routine.addComponent(self.target)

    def test_click_save_end_clickable_cases(self):
        """
        Test all combinations of options for what to save, what can be clicked on & what kind of clicks to end the
        routine on.
        """
        saveMouseStateCases = [
            {'val': "final",
             'want': ["thisExp.addData('testMouse.x', x)"],  # should contain code for adding final value of x
             'avoid': ["testMouse.x.append(x)"]},  # should not contain code to update testMouse.x in frame loop
            {'val': "on click",
             'want': ["thisExp.addData('testMouse.x', testMouse.x)",  # should add testMouse.x at the end
                      "testMouse.x.append(x)"],  # should contain code to update testMouse.x in frame loop
             'avoid': ["thisExp.addData('testMouse.x', x)"]},  # should not add final value of x
            {'val': "on valid click",
             'want': ["thisExp.addData('testMouse.x', testMouse.x)",  # should add testMouse.x at the end
                      "testMouse.x.append(x)",  # should contain code to update testMouse.x in frame loop
                      "if gotValidClick:"],  # should check for valid clicks
             'avoid': ["thisExp.addData('testMouse.x', x)"]},  # should not add final value of x
            {'val': "every frame",
             'want': ["thisExp.addData('testMouse.x', testMouse.x)",  # should add testMouse.x at the end
                      "testMouse.x.append(x)"],  # should contain code to update testMouse.x in frame loop
             'avoid': ["thisExp.addData('testMouse.x', x)"]},  # should not add final value of x
            {'val': "never",
             'want': [],
             'avoid': ["thisExp.addData('testMouse.x', testMouse.x)",  # should not add testMouse.x at the end
                       "testMouse.x.append(x)",  # should not contain code to update testMouse.x in frame loop
                       "thisExp.addData('testMouse.x', x)"]},  # should not add final value of x]},
        ]
        forceEndRoutineOnPressCases = [
            {'val':  "never",
             'want': [],
             'avoid': ["# abort routine on response",  # should not include code to abort routine
                       "# abort routine on response"]},
            {'val': "any click",
             'want': ["# abort routine on response"],  # should include code to abort routine on response
             'avoid': []},
            {'val': "valid click",
             'want': ["# abort routine on response",  # should include code to abort routine on response
                      "if gotValidClick:"],  # should check for valid clicks
             'avoid': []},
        ]
        clickableCases = [
            {'val': "[]",
             'want': [],
             'avoid': ["clickableList = [testPolygon]"]},  # should not define a populated clickables list
            {'val': "[testPolygon]",
             'want': [],
             'avoid': ["clickableList = []"]},  # should not define a blank clickables list
        ]
        # Iterate through saveMouseState cases
        for SMScase in saveMouseStateCases:
            # Set saveMouseState
            self.comp.params['saveMouseState'].val = SMScase['val']
            for FEROPcase in forceEndRoutineOnPressCases:
                # Set forceEndRoutineOnPress
                self.comp.params['forceEndRoutineOnPress'].val = FEROPcase['val']
                for Ccase in clickableCases:
                    # Set clickable
                    self.comp.params['clickable'].val = Ccase['val']

                    # Compile script
                    script = self.exp.writeScript(target="PsychoPy")
                    try:
                        # Look for wanted phrases
                        for phrase in SMScase['want']:
                            assert phrase in script, (
                                f"{phrase} not found in script when saveMouseState={self.comp.params['saveMouseState']}, "
                                f"forceEndRoutineOnPress={self.comp.params['forceEndRoutineOnPress']} and "
                                f"clickable={self.comp.params['clickable']}"
                            )
                        for phrase in FEROPcase['want']:
                            assert phrase in script, (
                                f"{phrase} not found in script when saveMouseState={self.comp.params['saveMouseState']}, "
                                f"forceEndRoutineOnPress={self.comp.params['forceEndRoutineOnPress']} and "
                                f"clickable={self.comp.params['clickable']}"
                            )
                        for phrase in Ccase['want']:
                            assert phrase in script, (
                                f"{phrase} not found in script when saveMouseState={self.comp.params['saveMouseState']}, "
                                f"forceEndRoutineOnPress={self.comp.params['forceEndRoutineOnPress']} and "
                                f"clickable={self.comp.params['clickable']}"
                            )
                        # Check there's no avoid phrases
                        for phrase in SMScase['avoid']:
                            assert phrase not in script, (
                                f"{phrase} found in script when saveMouseState={self.comp.params['saveMouseState']}, "
                                f"forceEndRoutineOnPress={self.comp.params['forceEndRoutineOnPress']} and "
                                f"clickable={self.comp.params['clickable']}"
                            )
                        for phrase in FEROPcase['avoid']:
                            assert phrase not in script, (
                                f"{phrase} found in script when saveMouseState={self.comp.params['saveMouseState']}, "
                                f"forceEndRoutineOnPress={self.comp.params['forceEndRoutineOnPress']} and "
                                f"clickable={self.comp.params['clickable']}"
                            )
                        for phrase in Ccase['avoid']:
                            assert phrase not in script, (
                                f"{phrase} found in script when saveMouseState={self.comp.params['saveMouseState']}, "
                                f"forceEndRoutineOnPress={self.comp.params['forceEndRoutineOnPress']} and "
                                f"clickable={self.comp.params['clickable']}"
                            )
                    except AssertionError as err:
                        # If any assertion fails, save script to view
                        filename = Path(TESTS_DATA_PATH) / f"{__class__.__name__}_clicks_cases_local.py"
                        with open(filename, "w") as f:
                            f.write(script)
                        # Append ref to saved script in error message
                        print(f"Script saved at: {filename}")
                        # Raise original error
                        raise err



from pathlib import Path

import psychopy.experiment
from psychopy.experiment.components.text import TextComponent
from psychopy.experiment._experiment import RequiredImport
from psychopy.tests.utils import TESTS_FONT
from os import path
import os, shutil, glob
import py_compile
import difflib
from tempfile import mkdtemp
import codecs
from psychopy import core, prefs
import pytest
import locale
import numpy
import sys
import re

import xmlschema

# Jeremy Gray March 2011

# caveats when comparing files:
# - dicts have no defined order, can load and save differently: use a
#   known-diff file to suppress boring errors.  This situation was
#   addressed in 7e2c72a for stimOut by sorting the keys
# - namespace.makeValid() can change var names from the orig demos,
#   but should not do so from a load-save-load because only the first
#   load should change things

allComponents = psychopy.experiment.getComponents(fetchIcons=False)
isTime = re.compile(r"\d+:\d+(:\d+)?( [AP]M)?")


def _filterout_legal(lines):
    """Ignore first 5 lines: header info, version, date can differ no problem
    """
    return [line
            for line in lines[5:]
            if not "This experiment was created using PsychoPy3 Experiment Builder (" in line and
            not ("trialList=data.importConditions(" in line and ".xlsx'))" in line)]
        #-This experiment was created using PsychoPy3 Experiment Builder (v1.65.01), August 03, 2011, at 13:14
        #+This experiment was created using PsychoPy3 Experiment Builder (v1.65.02), August 03, 2011, at 13:14
        #-    trialList=data.importConditions(u'trialTypes.xlsx'))
        #+    trialList=data.importConditions('trialTypes.xlsx'))
        #-    trialList=data.importConditions(u'mainTrials.xlsx'))
        #+    trialList=data.importConditions('mainTrials.xlsx'))

def _diff(a, b):
    """ diff of strings; returns a generator, 3 lines of context by default, - or + for changes """
    return difflib.unified_diff(a, b)

def _diff_file(a, b):
    """ diff of files read as strings, by line; output is similar to git gui diff """
    diff = _diff(open(a).readlines(), open(b).readlines())
    return list(diff)


class TestExpt():
    @classmethod
    def setup_class(cls):
        cls.exp = psychopy.experiment.Experiment() # create once, not every test
        try:
            cls.tmp_dir = mkdtemp(dir=Path(__file__).root, prefix='psychopy-tests-app')
        except (PermissionError, OSError):
            # can't write to root on Linux
            cls.tmp_dir = mkdtemp(prefix='psychopy-tests-app')

    def setup(self):
        """This setup is done for each test individually
        """
        self.here = path.abspath(path.dirname(__file__))
        self.known_diffs_file   = path.join(self.here, '../known_py_diffs.txt')
        self.tmp_diffs_file     = path.join(self.here, 'tmp_py_diffs.txt') # not deleted by mkdtemp cleanup

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.tmp_dir, ignore_errors=True)



    def _checkLoadSave(self, file):
        exp = self.exp
        py_file = file+'.py'
        psyexp_file = file+'newXML.psyexp'

        # go from psyexp file on disk to internal builder representation:
        self.exp.loadFromXML(file)
        self.exp.saveToXML(psyexp_file)
        assert len(self.exp.namespace.user) # should populate the namespace
        assert not self.exp.namespace.getCollisions() # ... without duplicates

        # generate a script, like 'lastrun.py':
        script = self.exp.writeScript()
        assert len(script) > 1500  # default empty script is ~2200 chars

        # save the script:
        with codecs.open(py_file, 'w', 'utf-8-sig') as f:
            f.write(script)

        return py_file, psyexp_file



    def _checkPyDiff(self, file_py, file2_py):
        """return '' for no meaningful diff, or a diff patch"""

        diff_py_lines = _diff_file(file_py, file2_py)[5:] # ignore first five lines --- +++
        if not len(diff_py_lines):
            return ''

        bad_diff = False # not all diffs are bad...
        # get all differing lines only, no context:
        f1 = _filterout_legal([x[1:] for x in diff_py_lines if x[0] == '+'])
        f2 = _filterout_legal([x[1:] for x in diff_py_lines if x[0] == '-'])
        if len(f1) != len(f2):
            bad_diff = True
        diff_py_keylines = f1 + f2

        # if line starts with stimOut, check for mere variation in order within line = ok
        #     fails for multiple stimOut lines with diff conditionss ==> len(set()) > 1
        if not bad_diff:
            # get only the stimOut lines, ignore leading whitespace:
            diff_py_stimOut = [y.replace("'", " ' ").strip() for y in diff_py_keylines
                                 if y.lstrip().startswith('stimOut')]
            # in Navon demo, stimOut comes from a dict written as a list, for conditions
            diff_py_stimOut_sort = []
            for line in diff_py_stimOut:
                sp_line = line.split()
                sp_line.sort() # squash order
                diff_py_stimOut_sort.append(' '.join(sp_line))
            # for diff lines that start with stimOut, are they same except order?
            if len(set(diff_py_stimOut_sort)) > 1: # set() squashes duplicates
                bad_diff = True
            # are the stimOut lines the only ones that differ? if so, we're ok
            if len(diff_py_keylines) != len(diff_py_stimOut):
                bad_diff = True

        # add another checks here:
        #if not bad_diff:
        #    some_condition = ...
        #    if some_condition:
        #        bad_diff = True

        # create a patch / diff file if bad_diff and not already a known-ok-diff:
        if bad_diff:
            diff_py_patch = ''.join(diff_py_lines)
            known_diffs = open(self.known_diffs_file).read()
            if known_diffs.find(diff_py_patch) < 0:
                patch = open(self.new_diff_file+'_'+path.basename(file_py)+'.patch', 'wb+')
                patch.write(path.basename(file_py) + ' load-save difference in resulting .py files: '
                            + '-'*15 + '\n' + diff_py_patch+'\n' + '-'*80 +'\n\n')
                patch.close()

                return diff_py_patch  # --> final assert will fail
        return ''

    def test_Exp_LoadCompilePsyexp(self):
        #""" for each builder demo .psyexp: load-save-load, compile (syntax check), namespace"""
        exp = self.exp
        self.new_diff_file = self.tmp_diffs_file

        # make temp copies of all builder demos:
        for root, dirs, files in os.walk(path.join(self.exp.prefsPaths['demos'], 'builder')):
            for f in files:
                if (f.endswith('.psyexp') or
                    f.endswith('.xlsx') or
                    f.endswith('.csv') )\
                    and not f.startswith('bart'):
                        shutil.copyfile(path.join(root, f), path.join(self.tmp_dir, f))
        # also copy any psyexp in 'here' (testExperiment dir)
        #for f in glob.glob(path.join(self.here, '*.psyexp')):
        #    shutil.copyfile(f, path.join(self.tmp_dir, path.basename(f)))
        test_psyexp = list(glob.glob(path.join(self.tmp_dir, '*.psyexp')))
        if len(test_psyexp) == 0:
            pytest.skip("No test .psyexp files found (no Builder demos??)")

        diff_in_file_py = '' # will later assert that this is empty
        #diff_in_file_psyexp = ''
        #diff_in_file_pyc = ''

        #savedLocale = '.'.join(locale.getlocale())
        locale.setlocale(locale.LC_ALL, '') # default
        if not sys.platform.startswith('win'):
            testlocList = ['en_US', 'en_US.UTF-8', 'ja_JP']
        else:
            testlocList = ['USA', 'JPN']

        for file in test_psyexp:
            # test for any diffs using various locale's:
            for loc in ['en_US', 'ja_JP']:
                try:
                    locale.setlocale(locale.LC_ALL, loc)
                except locale.Error:
                    continue #skip this locale; it isn't installed
                file_py, file_psyexp = self._checkLoadSave(file)
                file_pyc = self._checkCompile(file_py)
                #sha1_first = sha1hex(file_pyc, file=True)

                file2_py, file2_psyexp = self._checkLoadSave(file_psyexp)
                file2_pyc = self._checkCompile(file2_py)
                #sha1_second = sha1hex(file2_pyc, file=True)

                # check first against second, filtering out uninteresting diffs; catch diff in any of multiple psyexp files
                d = self._checkPyDiff(file_py, file2_py)
                if d:
                    diff_in_file_py += os.path.basename(file) + '::' + d
                #diff_psyexp = _diff_file(file_psyexp,file2_psyexp)[2:]
                #diff_in_file_psyexp += diff_psyexp
                #diff_pyc = (sha1_first != sha1_second)
                #assert not diff_pyc
        #locale.setlocale(locale.LC_ALL,'C')

        assert not diff_in_file_py ### see known_py_diffs.txt; potentially a locale issue? ###
        #assert not diff_in_file_psyexp # was failing most times, uninformative
        #assert not diff_in_file_pyc    # oops, was failing every time despite identical .py file

    def test_Run_FastStroopPsyExp(self):
        # start from a psyexp file, loadXML, execute, get keypresses from a emulator thread
        if sys.platform.startswith('linux'):
            pytest.skip("response emulation thread not working on linux yet")

        expfile = path.join(self.exp.prefsPaths['tests'], 'data', 'ghost_stroop.psyexp')
        with codecs.open(expfile, 'r', encoding='utf-8-sig') as f:
            text = f.read()

        # copy conditions file to tmp_dir
        shutil.copyfile(os.path.join(self.exp.prefsPaths['tests'], 'data', 'ghost_trialTypes.xlsx'),
                        os.path.join(self.tmp_dir,'ghost_trialTypes.xlsx'))

        # edit the file, to have a consistent font:
        text = text.replace("'Arial'", "'" + TESTS_FONT +"'")

        expfile = path.join(self.tmp_dir, 'ghost_stroop.psyexp')
        with codecs.open(expfile, 'w', encoding='utf-8-sig') as f:
            f.write(text)

        self.exp.loadFromXML(expfile)  # reload the edited file
        script = self.exp.writeScript()

        # reposition its window out from under splashscreen (can't do easily from .psyexp):
        script = script.replace('fullscr=False,','pos=(40,40), fullscr=False,')
        # Only log errors.
        script = script.replace('logging.console.setLevel(logging.WARNING',
                                'logging.console.setLevel(logging.ERROR')

        lastrun = path.join(self.tmp_dir, 'ghost_stroop_lastrun.py')
        with codecs.open(lastrun, 'w', encoding='utf-8-sig') as f:
            f.write(script)

        # run:
        stdout, stderr = core.shellCall([sys.executable, lastrun], stderr=True)
        assert not stderr

    def test_Exp_NameSpace(self):
        namespace = self.exp.namespace
        assert namespace.exists('psychopy') == "Psychopy module"

        namespace.add('foo')
        assert namespace.exists('foo') == "one of your Components, Routines, or condition parameters"
        namespace.add('foo')
        assert namespace.getCollisions() == ['foo']

        assert not namespace.isValid('123')
        assert not namespace.isValid('a1 23')
        assert not namespace.isValid('a123$')

        assert namespace.makeValid('123') == 'var_123'
        assert namespace.makeValid('123', prefix='wookie') == 'wookie_123'
        assert namespace.makeValid('a a a') == 'a_a_a'
        namespace.add('b')
        assert namespace.makeValid('b') == 'b_2'
        assert namespace.makeValid('a123$') == 'a123_'

        assert namespace.makeLoopIndex('trials') == 'thisTrial'
        assert namespace.makeLoopIndex('trials_2') == 'thisTrial_2'
        assert namespace.makeLoopIndex('stimuli') == 'thisStimulus'


class TestExpImports():
    def setup(self):
        self.exp = psychopy.experiment.Experiment()
        self.exp.requiredImports = []

    def test_requireImportName(self):
        import_ = RequiredImport(importName='foo', importFrom='',
                                 importAs='')
        self.exp.requireImport(importName='foo')

        assert import_ in self.exp.requiredImports

        script = self.exp.writeScript()
        assert 'import foo\n' in script

    def test_requireImportFrom(self):
        import_ = RequiredImport(importName='foo', importFrom='bar',
                                 importAs='')
        self.exp.requireImport(importName='foo', importFrom='bar')

        assert import_ in self.exp.requiredImports

        script = self.exp.writeScript()
        assert 'from bar import foo\n' in script

    def test_requireImportAs(self):
        import_ = RequiredImport(importName='foo', importFrom='',
                                 importAs='baz')
        self.exp.requireImport(importName='foo', importAs='baz')

        assert import_ in self.exp.requiredImports

        script = self.exp.writeScript()
        assert 'import foo as baz\n' in script

    def test_requireImportFromAs(self):
        import_ = RequiredImport(importName='foo', importFrom='bar',
                                 importAs='baz')
        self.exp.requireImport(importName='foo', importFrom='bar',
                               importAs='baz')

        assert import_ in self.exp.requiredImports

        script = self.exp.writeScript()
        assert 'from bar import foo as baz\n' in script

    def test_requirePsychopyLibs(self):
        import_ = RequiredImport(importName='foo', importFrom='psychopy',
                                 importAs='')
        self.exp.requirePsychopyLibs(['foo'])

        assert import_ in self.exp.requiredImports

        script = self.exp.writeScript()
        assert 'from psychopy import foo\n' in script

    def test_requirePsychopyLibs2(self):
        import_0 = RequiredImport(importName='foo', importFrom='psychopy',
                                  importAs='')
        import_1 = RequiredImport(importName='foo', importFrom='psychopy',
                                  importAs='')
        self.exp.requirePsychopyLibs(['foo', 'bar'])

        assert import_0 in self.exp.requiredImports
        assert import_1 in self.exp.requiredImports

        script = self.exp.writeScript()
        assert 'from psychopy import foo, bar\n' in script

    def test_requireImportAndPsychopyLib(self):
        import_0 = RequiredImport(importName='foo', importFrom='psychopy',
                                  importAs='')
        import_1 = RequiredImport(importName='bar', importFrom='',
                                  importAs='')
        self.exp.requirePsychopyLibs(['foo'])
        self.exp.requireImport('bar')

        assert import_0 in self.exp.requiredImports
        assert import_1 in self.exp.requiredImports

        script = self.exp.writeScript()
        print(script)
        assert 'from psychopy import foo\n' in script
        assert 'import bar\n' in script


class TestRunOnce():
    def setup(self):
        from psychopy.experiment import exports
        self.buff = exports.IndentingBuffer()

    def _doSingleTest(self, code):
        self.buff.writeOnceIndentedLines(code)
        assert code in self.buff._writtenOnce
        script = self.buff.getvalue()
        assert code + '\n' in script

    def test_runOnceSingleLine(self):
        self._doSingleTest('foo bar baz')

    def test_runOnceMultiLine(self):
        self._doSingleTest('foo bar baz\nbla bla bla')

    def test_runOnceMultipleStatements(self):
        code_0 = 'foo bar code0'
        self.buff.writeOnceIndentedLines(code_0)
        code_1 = 'bla bla code1'
        self.buff.writeOnceIndentedLines(code_1)

        assert code_0 in self.buff._writtenOnce
        assert code_1 in self.buff._writtenOnce
        script = self.buff.getvalue()
        assert (code_0 + '\n' + code_1 + '\n') in script

    def test_repeatedInsert(self):
        code = 'uniqueBlah'
        self.buff.writeOnceIndentedLines(code)  # write it twice
        self.buff.writeOnceIndentedLines(code)
        # check it appears only once
        script = self.buff.getvalue()
        assert script.count(code) == 1



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import math
from psychopy.iohub.constants import EventConstants, EyeTrackerConstants
from psychopy.iohub.devices import Computer, Device
from psychopy.iohub.devices.eyetracker import EyeTrackerDevice
from psychopy.iohub.devices.eyetracker.hw.tobii.tobiiCalibrationGraphics import TobiiPsychopyCalibrationGraphics
from psychopy.iohub.devices.eyetracker.eye_events import *
from psychopy.iohub.errors import print2err, printExceptionDetailsToStdErr
try:
    from .tobiiwrapper import TobiiTracker
except Exception:
    print2err('Error importing tobiiwrapper.TobiiTracker')
    printExceptionDetailsToStdErr()


class EyeTracker(EyeTrackerDevice):
    """
    To start iohub with a Tobii eye tracker device, add the Tobii
    device to the dictionary passed to launchHubServer or the 
    experiment's iohub_config.yaml::

        eyetracker.hw.tobii.EyeTracker
        
    Examples:
        A. Start ioHub with a Tobii device and run tracker calibration::
    
            from psychopy.iohub import launchHubServer
            from psychopy.core import getTime, wait

            iohub_config = {'eyetracker.hw.tobii.EyeTracker':
                {'name': 'tracker', 'runtime_settings': {'sampling_rate': 120}}}
                
            io = launchHubServer(**iohub_config)
            
            # Get the eye tracker device.
            tracker = io.devices.tracker
                            
            # run eyetracker calibration
            r = tracker.runSetupProcedure()
            
        B. Print all eye tracker events received for 2 seconds::
                        
            # Check for and print any eye tracker events received...
            tracker.setRecordingState(True)
            
            stime = getTime()
            while getTime()-stime < 2.0:
                for e in tracker.getEvents():
                    print(e)
            
        C. Print current eye position for 5 seconds::
                        
            # Check for and print current eye position every 100 msec.
            stime = getTime()
            while getTime()-stime < 5.0:
                print(tracker.getPosition())
                wait(0.1)
            
            tracker.setRecordingState(False)
            
            # Stop the ioHub Server
            io.quit()
    """
    _tobii = None

    DEVICE_TIMEBASE_TO_SEC = 0.000001
    EVENT_CLASS_NAMES = [
        'MonocularEyeSampleEvent',
        'BinocularEyeSampleEvent',
        'FixationStartEvent',
        'FixationEndEvent',
        'SaccadeStartEvent',
        'SaccadeEndEvent',
        'BlinkStartEvent',
        'BlinkEndEvent']
    __slots__ = []

    def __init__(self, *args, **kwargs):
        EyeTrackerDevice.__init__(self, *args, **kwargs)

        if self.model_name:
            self.model_name = self.model_name.strip()
            if len(self.model_name) == 0:
                self.model_name = None
        model_name = self.model_name
        serial_num = self.getConfiguration().get('serial_number')

        EyeTracker._tobii = None
        try:
            EyeTracker._tobii = TobiiTracker(serial_num, model_name)
        except Exception:
            print2err('Error creating Tobii Device class')
            printExceptionDetailsToStdErr()

        # Apply license file if needed
        try:
            license_file = self.getConfiguration().get('license_file', "")
            if license_file != "":
                with open(license_file, "rb") as f:
                    license = f.read()
                    res = self._tobii._eyetracker.apply_licenses(license)
                    if len(res) == 0:
                        print2err("Successfully applied Tobii license from: {}".format(license_file))
                    else:
                        print2err("Error: Failed to apply Tobii license from single key. "
                                  "Validation result: %s." % (res[0].validation_result))
            else:
                print2err("No Tobii license_file in config. Skipping.")
        except Exception:
            print2err("Error calling Tobii.apply_licenses with file {}.".format(license_file))
            printExceptionDetailsToStdErr()

        srate = self._runtime_settings['sampling_rate']
        if srate and srate in self._tobii.getAvailableSamplingRates():
            self._tobii.setSamplingRate(srate)

        self._latest_sample = None
        self._latest_gaze_position = None

    def trackerTime(self):
        """Current eye tracker time in the eye tracker's native time base. The
        Tobii system uses a usec timebase.

        Args:
            None

        Returns:
            float: current native eye tracker time. (in usec for the Tobii)

        """
        if self._tobii:
            return self._tobii.getCurrentEyeTrackerTime()
        return EyeTrackerConstants.EYETRACKER_ERROR

    def trackerSec(self):
        """Current eye tracker time, normalized to sec.msec format.

        Args:
            None

        Returns:
            float: current native eye tracker time in sec.msec-usec format.

        """
        if self._tobii:
            return self._tobii.getCurrentEyeTrackerTime() * self.DEVICE_TIMEBASE_TO_SEC
        return EyeTrackerConstants.EYETRACKER_ERROR

    def setConnectionState(self, enable):
        """
        setConnectionState is a no-op when using the Tobii system, as the
        connection is established when the Tobii EyeTracker classes are created,
        and remains active until the program ends, or a error occurs resulting
        in the loss of the tracker connection.

        Args:
            enable (bool): True = enable the connection, False = disable the connection.

        Return:
            bool: indicates the current connection state to the eye tracking hardware.
        """
        if self._tobii:
            return True
        return False
    
    def isConnected(self):
        """isConnected returns whether the Tobii is connected to the experiment
        PC and if the tracker state is valid. Returns True if the tracker can
        be put into Record mode, etc and False if there is an error with the
        tracker or tracker connection with the experiment PC.

        Args:
            None

        Return:
            bool:  True = the eye tracking hardware is connected. False otherwise.

        """
        if self._tobii:
            return True
        return False

    def sendMessage(self, message_contents, time_offset=None):
        """The sendMessage method is not supported by the Tobii implementation
        of the Common Eye Tracker Interface, as the Tobii SDK does not support
        saving eye data to a native data file during recording."""
        return EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED

    def sendCommand(self, key, value=None):
        """The sendCommand method is not supported by the Tobii Common Eye
        Tracker Interface."""
        return EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED

    def runSetupProcedure(self, calibration_args={}):
        """runSetupProcedure performs a calibration routine for the Tobii eye
        tracking system.
        """
        try:
            genv = TobiiPsychopyCalibrationGraphics(self, calibration_args)

            calibrationOK = genv.runCalibration()

            # On some graphics cards, we have to minimize before closing or the calibration window will stay visible
            # after close is called.
            genv.window.winHandle.set_visible(False)
            genv.window.winHandle.minimize()

            genv.window.close()

            genv._unregisterEventMonitors()
            genv.clearAllEventBuffers()

            return calibrationOK

        except Exception:
            print2err('Error during runSetupProcedure')
            printExceptionDetailsToStdErr()
        return EyeTrackerConstants.EYETRACKER_ERROR

    def enableEventReporting(self, enabled=True):
        """enableEventReporting is functionally identical to the eye tracker
        device specific enableEventReporting method."""

        try:
            enabled = EyeTrackerDevice.enableEventReporting(self, enabled)
            self.setRecordingState(enabled)
            return enabled
        except Exception as e:
            print2err('Error during enableEventReporting')
            printExceptionDetailsToStdErr()
        return EyeTrackerConstants.EYETRACKER_ERROR

    def setRecordingState(self, recording):
        """setRecordingState is used to start or stop the recording of data
        from the eye tracking device.

        args:
           recording (bool): if True, the eye tracker will start recordng available
              eye data and sending it to the experiment program if data streaming
              was enabled for the device. If recording == False, then the eye
              tracker stops recording eye data and streaming it to the experiment.

        If the eye tracker is already recording, and setRecordingState(True) is
        called, the eye tracker will simple continue recording and the method call
        is a no-op. Likewise if the system has already stopped recording and
        setRecordingState(False) is called again.

        Args:
            recording (bool): if True, the eye tracker will start recordng data.; false = stop recording data.

        Return:
            bool: the current recording state of the eye tracking device

        """
        if self._tobii and recording is True and not self.isRecordingEnabled():
            #ioHub.print2err("Starting Tracking... ")
            self._tobii.startTracking(self._handleNativeEvent)
            return EyeTrackerDevice.enableEventReporting(self, True)

        elif self._tobii and recording is False and self.isRecordingEnabled():
            self._tobii.stopTracking()
            #ioHub.print2err("Stopping Tracking... ")
            self._latest_sample = None
            self._latest_gaze_position = None
            return EyeTrackerDevice.enableEventReporting(self, False)

        return self.isRecordingEnabled()

    def isRecordingEnabled(self):
        """isRecordingEnabled returns the recording state from the eye tracking
        device.

        Args:
           None

        Return:
            bool: True == the device is recording data; False == Recording is not occurring

        """
        if self._tobii:
            return self._tobii._isRecording
        return False

    def getLastSample(self):
        """Returns the latest sample retrieved from the Tobii device. The Tobii
        system always using the BinocularSample Event type.

        Args:
            None

        Returns:
            None: If the eye tracker is not currently recording data.

            EyeSample: If the eye tracker is recording in a monocular tracking mode, the latest sample event of this event type is returned.

            BinocularEyeSample:  If the eye tracker is recording in a binocular tracking mode, the latest sample event of this event type is returned.

        """
        return self._latest_sample

    def getLastGazePosition(self):
        """Returns the latest 2D eye gaze position retrieved from the Tobii
        device. This represents where the eye tracker is reporting each eye
        gaze vector is intersecting the calibrated surface.

        In general, the y or vertical component of each eyes gaze position should
        be the same value, since in typical user populations the two eyes are
        yoked vertically when they move. Therefore any difference between the
        two eyes in the y dimension is likely due to eye tracker error.

        Differences between the x, or horizontal component of the gaze position,
        indicate that the participant is being reported as looking behind or
        in front of the calibrated plane. When a user is looking at the
        calibration surface , the x component of the two eyes gaze position should be the same.
        Differences between the x value for each eye either indicates that the
        user is not focussing at the calibrated depth, or that there is error in the eye data.

        The above remarks are true for any eye tracker in general.

        The getLastGazePosition method returns the most recent eye gaze position
        retrieved from the eye tracker device. This is the position on the
        calibrated 2D surface that the eye tracker is reporting as the current
        eye position. The units are in the units in use by the Display device.

        If binocular recording is being performed, the average position of both
        eyes is returned.

        If no samples have been received from the eye tracker, or the
        eye tracker is not currently recording data, None is returned.

        Args:
            None

        Returns:
            None: If the eye tracker is not currently recording data or no eye samples have been received.

            tuple: Latest (gaze_x,gaze_y) position of the eye(s)

        """
        return self._latest_gaze_position

    def _setSamplingRate(self, sampling_rate):
        return self._tobii.setSamplingRate(sampling_rate)

    def _poll(self):
        """The Tobii system uses a callback approach to providing new eye data
        as it becomes available, so polling (and therefore this method) are not
        used."""
        pass

    def _handleNativeEvent(self, *args, **kwargs):
        """This method is called every time there is new eye data available
        from the Tobii system, which will be roughly equal to the sampling rate
        eye data is being recorded at.

        The callback needs to return as quickly as possible so there is
        no chance of overlapping calls being made to the callback.
        Therefore this method simply puts the event data received from
        the eye tracker device, and the local ioHub time the callback
        was called, into a buffer for processing by the ioHub event
        system.
        """
        if self.isReportingEvents():
            try:
                logged_time = Computer.getTime()
                tobii_logged_time = self._tobii.getCurrentLocalTobiiTime() * self.DEVICE_TIMEBASE_TO_SEC
                
                eye_data_event = args[0]
                
                data_delay = tobii_logged_time - (eye_data_event['system_time_stamp'] * self.DEVICE_TIMEBASE_TO_SEC)
        
                device_event_time = eye_data_event['device_time_stamp']
                iohub_event_time = (logged_time - data_delay)
                self._addNativeEventToBuffer(
                    (logged_time,
                     device_event_time,
                     iohub_event_time,
                     data_delay,
                     eye_data_event))
                return True
            except Exception:
                print2err('ERROR IN _handleNativeEvent')
                printExceptionDetailsToStdErr()
        else:
            print2err(
                'self._handleNativeEvent called but isReportingEvents == false')


    def _getIOHubEventObject(self, native_event_data):
        """The _getIOHubEventObject method is called by the ioHub Server to
        convert new native device event objects that have been received to the
        appropriate ioHub Event type representation.

        The Tobii ioHub eye tracker implementation uses a callback method
        to register new native device events with the ioHub Server.
        Therefore this method converts the native Tobii event data into
        an appropriate ioHub Event representation.

        Args:
            native_event_data: object or tuple of (callback_time, native_event_object)

        Returns:
            tuple: The appropriate ioHub Event type in list form.

        """
        try:
            logged_time, device_event_time, iohub_event_time, data_delay, eye_data_event = native_event_data

            event_type = EventConstants.BINOCULAR_EYE_SAMPLE
    
            left_gaze_x, left_gaze_y = eye_data_event['left_gaze_point_on_display_area']
            right_gaze_x, right_gaze_y = eye_data_event['right_gaze_point_on_display_area']
    
            status = 0

            if eye_data_event['left_gaze_point_validity'] > 0:
                left_gaze_x, left_gaze_y = self._eyeTrackerToDisplayCoords(
                    (left_gaze_x, left_gaze_y))
            else:
                status += 20
                
            if eye_data_event['right_gaze_point_validity'] > 0:
                right_gaze_x, right_gaze_y = self._eyeTrackerToDisplayCoords(
                    (right_gaze_x, right_gaze_y))
            else:
                status += 2                

            right_gx, right_gy, right_gz = eye_data_event['right_gaze_origin_in_trackbox_coordinate_system']
            left_gx, left_gy, left_gz = eye_data_event['left_gaze_origin_in_trackbox_coordinate_system']

            confidenceInterval = 0.0
            binocSample = [
                0,
                0,
                0,  # device id (not currently used)
                Device._getNextEventID(),
                event_type,
                device_event_time,
                logged_time,
                iohub_event_time,
                confidenceInterval,
                data_delay,
                0,  # filtered id (always 0 right now)
                left_gaze_x,
                left_gaze_y,
                EyeTrackerConstants.UNDEFINED,
                left_gx,
                left_gy,
                left_gz,
                EyeTrackerConstants.UNDEFINED,  # Left Eye Angle x
                EyeTrackerConstants.UNDEFINED,  # Left Eye Angle y
                EyeTrackerConstants.UNDEFINED,  # Left Camera Sensor position x
                EyeTrackerConstants.UNDEFINED,  # Left Camera Sensor position y
                eye_data_event['left_pupil_diameter'],
                EyeTrackerConstants.PUPIL_DIAMETER_MM,
                EyeTrackerConstants.UNDEFINED,  # Left pupil size measure 2
                EyeTrackerConstants.UNDEFINED,  # Left pupil size measure 2 type
                EyeTrackerConstants.UNDEFINED,  # Left PPD x
                EyeTrackerConstants.UNDEFINED,  # Left PPD y
                EyeTrackerConstants.UNDEFINED,  # Left velocity x
                EyeTrackerConstants.UNDEFINED,  # Left velocity y
                EyeTrackerConstants.UNDEFINED,  # Left velocity xy
                right_gaze_x,
                right_gaze_y,
                EyeTrackerConstants.UNDEFINED,  # Right Eye Angle z
                right_gx,
                right_gy,
                right_gz,
                EyeTrackerConstants.UNDEFINED,  # Right Eye Angle x
                EyeTrackerConstants.UNDEFINED,  # Right Eye Angle y
                EyeTrackerConstants.UNDEFINED,  # Right Camera Sensor position x
                EyeTrackerConstants.UNDEFINED,  # Right Camera Sensor position y
                eye_data_event['right_pupil_diameter'],
                EyeTrackerConstants.PUPIL_DIAMETER_MM,
                EyeTrackerConstants.UNDEFINED,  # Right pupil size measure 2
                EyeTrackerConstants.UNDEFINED,  # Right pupil size measure 2 type
                EyeTrackerConstants.UNDEFINED,  # Right PPD x
                EyeTrackerConstants.UNDEFINED,  # Right PPD y
                EyeTrackerConstants.UNDEFINED,  # right velocity x
                EyeTrackerConstants.UNDEFINED,  # right velocity y
                EyeTrackerConstants.UNDEFINED,  # right velocity xy
                status
            ]
    
            self._latest_sample = binocSample
    
            if eye_data_event['left_gaze_point_validity'] == eye_data_event['right_gaze_point_validity'] == 0:
                self._latest_gaze_position = None
            elif eye_data_event['left_gaze_point_validity'] == eye_data_event['right_gaze_point_validity'] == 1:
                self._latest_gaze_position = [(right_gaze_x + left_gaze_x) / 2.0,
                                              (right_gaze_y + left_gaze_y) / 2.0]
            elif eye_data_event['left_gaze_point_validity'] == 1:
                self._latest_gaze_position = [left_gaze_x, left_gaze_y]
            elif eye_data_event['right_gaze_point_validity'] == 1:
                self._latest_gaze_position = [right_gaze_x, right_gaze_y]
    
            self._last_callback_time = logged_time
    
            return binocSample
        except Exception:
            printExceptionDetailsToStdErr()
        return None

    def _eyeTrackerToDisplayCoords(self, eyetracker_point):
        """Converts Tobii gaze positions to the Display device coordinate
        space."""
        gaze_x, gaze_y = eyetracker_point
        left, top, right, bottom = self._display_device.getCoordBounds()
        w, h = right - left, top - bottom
        x, y = left + w * gaze_x, bottom + h * (1.0 - gaze_y)
        return x, y

    def _displayToEyeTrackerCoords(self, display_x, display_y):
        """Converts a Display device point to Tobii gaze position coordinate
        space."""
        left, top, right, bottom = self._display_device.getCoordBounds()
        w, h = right - left, top - bottom

        return (left - display_x) / w, (top - display_y) / h

    def _close(self):
        if EyeTracker._tobii:
            EyeTracker._tobii.disconnect()
            EyeTracker._tobii = None
        EyeTrackerDevice._close(self)



"""Test PsychoPy sound.py using pygame backend; will fail if have already used pyo
"""

from psychopy import prefs, core
prefs.hardware['audioLib'] = ['pygame']

import pytest
import shutil
from tempfile import mkdtemp
from psychopy import sound #, microphone

#import pyo
import numpy

# py.test --cov-report term-missing --cov sound.py tests/test_sound/test_sound_pygame.py

from psychopy.tests.utils import TESTS_PATH, TESTS_DATA_PATH

@pytest.mark.needs_sound
class TestPygame:
    @classmethod
    def setup_class(self):
        self.contextName='pyo'
        try:
            assert sound.Sound == sound.SoundPygame
        except Exception:
            pytest.xfail('need to be using pygame')
        self.tmp = mkdtemp(prefix='psychopy-tests-sound')

    @classmethod
    def teardown_class(self):
        if hasattr(self, 'tmp'):
            shutil.rmtree(self.tmp, ignore_errors=True)

    def test_init(self):
        for note in ['A', 440, '440', [1,2,3,4], numpy.array([1,2,3,4])]:
            sound.Sound(note, secs=.1)
        with pytest.raises(ValueError):
            sound.Sound('this is not a file name')
        with pytest.raises(ValueError):
            sound.Sound(-1)
        with pytest.raises(ValueError):
            sound.Sound(440, secs=-1)
        with pytest.raises(ValueError):
            sound.Sound(440, secs=0)
        with pytest.raises(DeprecationWarning):
            sound.setaudioLib('foo')

        points = 100
        snd = numpy.ones(points) / 20

        #testFile = os.path.join(self.tmp, 'green_48000.wav')
        #r, d = wavfile.read(testFile)
        #assert r == 48000
        #assert len(d) == 92160
        #s = sound.Sound(testFile)

    def test_play(self):
        s = sound.Sound(secs=0.1)
        s.play()
        core.wait(s.getDuration()+.1)  # allows coverage of _onEOS
        s.play(loops=1)
        core.wait(s.getDuration()*2+.1)
        s.play(loops=-1)
        s.stop()

    def test_methods(self):
        s = sound.Sound(secs=0.1)
        v = s.getVolume()
        assert v == 1
        assert s.setVolume(0.5) == 0.5
        #assert s.setLoops(2) == 2
